"""
    created 15.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import unittest
from pathlib import Path

from for_runners.gpx import parse_gpx_file
from for_runners.svg import gpx2svg_string

BASE_PATH = Path(__file__).parent


class SvgTest(unittest.TestCase):

    def test_svg(self):
        filepath = Path(BASE_PATH, "fixture_files/parliament_buildings.gpx")

        gpxpy_instance = parse_gpx_file(filepath)
        svg_string = gpx2svg_string(gpxpy_instance, pretty=True)
        print("-" * 79)
        print(svg_string)
        print("-" * 79)

        svg_reference = Path(BASE_PATH, "fixture_files/parliament_buildings.svg")
        with svg_reference.open("r") as f:
            svg_reference_string = f.read()

        self.assertEqual(svg_string, svg_reference_string)



# Generated by Django 2.0.7 on 2018-07-10 19:10

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('for_runners', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='CostModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('createtime', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='Create time')),
                ('lastupdatetime', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='Time of the last change.')),
                ('name', models.CharField(help_text='Your start number', max_length=15)),
                ('amount', models.DecimalField(decimal_places=2, help_text='How much did you pay for this?', max_digits=8)),
            ],
            options={
                'verbose_name': 'Participation Cost',
                'verbose_name_plural': 'Participation Costs',
            },
        ),
        migrations.CreateModel(
            name='ParticipationModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('createtime', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='Create time')),
                ('lastupdatetime', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='Time of the last change.')),
                ('duration', models.TimeField(blank=True, help_text='You officially measured finisher time', null=True, verbose_name='Duration')),
                ('start_number', models.CharField(blank=True, help_text='Your start number', max_length=15, null=True)),
                ('finisher_count', models.PositiveIntegerField(blank=True, help_text='Number of participants who have finished in your discipline', null=True)),
                ('distance', models.ForeignKey(help_text='Length in meters of your event participation', on_delete=django.db.models.deletion.CASCADE, related_name='event_participations', to='for_runners.DistanceModel')),
                ('event', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='participations', to='for_runners.EventModel')),
                ('person', models.ForeignKey(help_text='The person who participated to this competition.', on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Event Participation',
                'verbose_name_plural': 'Event Participations',
            },
        ),
        migrations.AddField(
            model_name='costmodel',
            name='participation',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='costs', to='for_runners.ParticipationModel'),
        ),
    ]



"""
    created 31.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import collections
import logging

from geopy.geocoders import Nominatim

log = logging.getLogger(__name__)
Address = collections.namedtuple('Address', ('short, full'))


def construct_short_address(address):
    # from pprint import pprint
    # pprint(address)
    parts = []

    if "village" in address and "county" in address:
        parts.append("%s," % address["village"])
        parts.append(address["county"])
    else:
        parts.append(address.get("city") or address.get("town") or address.get("county") or address.get("state"))
        parts.append(address.get("suburb"))

    short_address = " ".join([part for part in parts if part])
    return short_address


def reverse_geo(lat, lon):
    """
    Create short+full address string from given geo coordinates

    >>> address = reverse_geo("51.6239133", "6.9749074")
    >>> address.short
    'Feldhausen, Bottrop'
    >>> address.full
    'Movie Park Germany, 1, Warner - Allee, Kuhberg, Feldhausen, Bottrop, Regierungsbezirk Münster, Nordrhein-Westfalen, 46244, Deutschland'

    >>> reverse_geo("52.518611", "13.376111").short
    'Berlin Tiergarten'

    :return: Address named tuple
        short : string
            The "sort" Address
        full : string
            The "full" Address
    """
    geolocator = Nominatim()
    location = geolocator.reverse("%s, %s" % (lat, lon))

    short_address = construct_short_address(address=location.raw["address"])

    return Address(short_address, location.address)


if __name__ == "__main__":
    print(reverse_geo("51.6239133", "6.9749074")) # Bottrop Feldhausen



"""
    created 12.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from pathlib import Path

from django.test import TestCase
from django_tools.unittest_utils.user import TestUserMixin
# https://github.com/jedie/django-for-runners
from for_runners.models import GpxModel

BASE_PATH = Path(__file__).parent


class GpxTests(TestUserMixin, TestCase):

    def test_parse_gpx(self):

        user = self._get_user(usertype="normal")

        filepath = Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx")
        with filepath.open("r") as f:
            gpx_content = f.read()

        instance = GpxModel.objects.add_gpx(gpx_content, user)
        print(instance)

        self.assertEqual(repr(instance), "<GpxModel: 2018-02-21 Moers Hülsdonk>")

        self.assertEqual(instance.points_no, 3)
        self.assertEqual(round(instance.length, 3), 4.727)
        self.assertEqual(instance.duration_s, 2)
        self.assertEqual(round(instance.pace, 3), 7.052)
        self.assertEqual(instance.heart_rate_avg, 125)

        self.assertEqual(instance.get_short_slug(), "2018-02-21-moers-hulsdonk")



"""
    created 21.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import unittest


class BaseTestCase(unittest.TestCase):

    def assert_equal_rounded(self, value1, value2, decimal_places=5, msg=None):

        if isinstance(value1, (tuple, list)):
            value1 = [round(v, decimal_places) for v in value1]
            value2 = [round(v, decimal_places) for v in value2]
        else:
            value1 = round(value1, decimal_places)
            value2 = round(value2, decimal_places)
        self.assertEqual(value1, value2, msg=msg)



"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import collections
import logging
from pprint import pprint

from django.conf import settings
from django.contrib import admin
from django.contrib.admin.views.main import ChangeList
from django.db import models
from django.template.response import TemplateResponse
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

# https://github.com/jedie/django-for-runners
from for_runners.admin.utils import BaseChangelistView
from for_runners.gpx_tools.humanize import human_distance, human_seconds, human_duration, convert_cash_values
from for_runners.models import EventLinkModel, EventModel
from for_runners.models.event import CostModel, ParticipationModel

log = logging.getLogger(__name__)


@admin.register(EventLinkModel)
class EventLinkModelAdmin(admin.ModelAdmin):
    pass


class LinkModelInline(admin.TabularInline):
    model = EventLinkModel
    extra = 2
    min_num = 0
    max_num = None
    fields = (
        'url',
        'text',
        'title',
    )


class HasTracksFilter(admin.SimpleListFilter):
    title = _('has GPX tracks')
    parameter_name = "tracks"

    def lookups(self, request, model_admin):
        return (
            ('y', _('yes')),
            ('n', _('no')),
        )

    def queryset(self, request, queryset):
        if self.value() == 'y':
            return queryset.exclude(participations__tracks__isnull=True)  # FIXME!
        if self.value() == 'n':
            return queryset.filter(participations__tracks__isnull=True)  # FIXME!


class EventStatistics:
    KEY_DISTANCE = "_distance"
    KEY_DURATION = "_duration"

    def __init__(self):
        self.convert_map = {
            self.KEY_DISTANCE: human_distance,
            self.KEY_DURATION: human_duration,
        }

    def get(self, events):
        raw_person_data = collections.defaultdict(collections.Counter)

        for event in events:
            # print(event)
            for participation in event.participations.all():
                # print(participation.person)
                participation_data = {
                    "_count": 1,
                    self.KEY_DISTANCE: participation.distance_km or 0,
                    self.KEY_DURATION: participation.get_duration_s() or 0,
                }

                for cost in participation.costs.all():
                    # print(cost)
                    participation_data[cost.name] = cost.amount

                raw_person_data[participation.person.username] += collections.Counter(participation_data)

        # turn defaultdict and counter to normal dict and convert some values:

        total_costs = collections.Counter()

        # pprint(raw_person_data)
        person_data = {}
        for username, counter_data in raw_person_data.items():
            user_data = []
            counter_data = dict(counter_data)
            for key, value in sorted(counter_data.items()):
                try:
                    func = self.convert_map[key]
                except KeyError:
                    pass
                else:
                    value = func(value)

                if key.startswith("_"):
                    # internal key
                    key = key[1:]
                else:
                    # costs item
                    # collect total costs:
                    total_costs += collections.Counter({"total": value, key: value})
                    # convert decimal field:
                    value = convert_cash_values(value)

                user_data.append((key, value))
            person_data[username] = user_data

        total_costs = dict(total_costs)
        total = total_costs.pop("total") # add total as lates item
        total_costs = [(name, convert_cash_values(amount)) for name, amount in sorted(total_costs.items())]
        total_costs.append(
            (_("total"), convert_cash_values(total))
        )

        # pprint(person_data)
        # pprint(total_costs)
        return person_data, total_costs


class StatisticsView(BaseChangelistView):

    template_name = "admin/for_runners/eventmodel/statistics.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        events = self.change_list.queryset  # filtered EventModel queryset form GpxModelChangeList
        # print("Events:", events.count())

        person_data, total_costs = EventStatistics().get(events)

        context.update({
            "title": _("Event Statistics"),
            "person_data": person_data,
            "total_costs": total_costs,
            "user": self.request.user,
            "opts": EventModel._meta,
        })
        return context


class EventModelChangeList(ChangeList):

    def get_results(self, request):
        super().get_results(request)

        view = StatisticsView.as_view()
        response = view(request, self)
        assert isinstance(response, TemplateResponse), "Method %s didn't return a TemplateResponse!" % view
        self.statistics = response.rendered_content


@admin.register(EventModel)
class EventModelAdmin(admin.ModelAdmin):

    def participations(self, obj):
        html = []
        participations = obj.participations.all()
        for participation in participations:
            html.append("%s (%s)" % (
                participation.person.username,
                participation.get_human_distance(),
            ))
        html = "<br>".join(html)
        html = mark_safe(html)
        return html
    participations.short_description = _("Participations")

    list_display = ("verbose_name", "participations", "links_html", "start_date", "discipline")
    date_hierarchy = "start_date"
    list_filter = (HasTracksFilter, "participations__person")
    list_display_links = ("verbose_name",)
    search_fields = ("name",)
    inlines = [
        LinkModelInline,
    ]

    def get_changelist(self, request, **kwargs):
        """
        Returns the ChangeList class for use on the changelist page.
        """
        return EventModelChangeList


class CostModelInline(admin.TabularInline):
    model = CostModel
    extra = 2
    min_num = 0
    max_num = None


@admin.register(ParticipationModel)
class ParticipationModelAdmin(admin.ModelAdmin):

    def human_duration(self, obj):
        if obj.duration:
            return human_seconds(obj.get_duration_s())
    human_duration.short_description = _("Duration")
    human_duration.admin_order_field = "duration"

    def pace(self, obj):
        pace_s = obj.get_pace_s()
        if pace_s:
            return "%s min/km" % human_seconds(pace_s)
    pace.short_description = _("Pace")

    list_display = ("event", "distance_km", "finisher_count", "person", "start_number", "human_duration", "pace")
    list_filter = ("person", "distance_km")
    date_hierarchy = "event__start_date"
    search_fields = (
        "person__username",
        "event__name",
        "event__start_date",
    )
    inlines = [
        CostModelInline,
    ]



"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging

from django.db import models
from django.utils.translation import ugettext_lazy as _

log = logging.getLogger(__name__)


class DisciplineModel(models.Model):
    name = models.CharField(max_length=255, help_text=_("Sport discipline"))

    def __str__(self):
        return self.name



"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import datetime
import decimal
import math

from django.conf import settings
from django.utils.translation import ugettext_lazy as _


def human_seconds(duration):
    """
    >>> human_seconds(1)
    '0:01'
    >>> human_seconds(61)
    '1:01'
    >>> human_seconds(60*60+1)
    '1:00:01'
    """
    parts = []

    hours = None
    if duration>1*60*60:
        hours = math.floor(duration/60/60)
        duration -= hours*60*60
        parts.append("%i" % hours)

    minutes = math.floor(duration/60)
    if hours is None:
        parts.append("%i" % minutes)
    else:
        parts.append("%02i" % minutes)

    seconds = math.floor(duration - (minutes*60))
    parts.append("%02i" % seconds)

    return ":".join(parts)

def human_duration(t):
    """
    Converts a time duration into a friendly text representation.

    >>> human_duration("type error")
    Traceback (most recent call last):
        ...
    decimal.InvalidOperation: [<class 'decimal.ConversionSyntax'>]


    >>> human_duration(datetime.timedelta(microseconds=1000))
    '1.0 ms'
    >>> human_duration(0.01)
    '10.0 ms'
    >>> human_duration(0.9)
    '900.0 ms'
    >>> human_duration(datetime.timedelta(seconds=1))
    '1.0 sec'
    >>> human_duration(65.5)
    '1.1 min'
    >>> human_duration(59.1 * 60)
    '59.1 min'
    >>> human_duration(60*60)
    '1.0 hours'
    >>> human_duration(1.06*60*60)
    '1.1 hours'
    >>> human_duration(datetime.timedelta(hours=24))
    '1.0 days'
    >>> human_duration(2.54 * 60 * 60 * 24 * 365)
    '2.5 years'
    """
    if isinstance(t, datetime.timedelta):
        t = decimal.Decimal(t.total_seconds())
    elif not isinstance(t, decimal.Decimal):
        t = decimal.Decimal(t)

    chunks = (
      (decimal.Decimal(60 * 60 * 24 * 365), _('years')),
      (decimal.Decimal(60 * 60 * 24 * 30), _('months')),
      (decimal.Decimal(60 * 60 * 24 * 7), _('weeks')),
      (decimal.Decimal(60 * 60 * 24), _('days')),
      (decimal.Decimal(60 * 60), _('hours')),
    )

    if t < 1:
        return _("%.1f ms") % round(t * 1000, 1)
    if t < 60:
        return _("%.1f sec") % round(t, 1)
    if t < 60 * 60:
        return _("%.1f min") % round(t / 60, 1)

    for seconds, name in chunks:
        count = t / seconds
        if count >= 1:
            count = round(count, 1)
            break
    return "%(number).1f %(type)s" % {'number': count, 'type': name}


def human_distance(km):
    """
    >>> human_distance(km=0.10000001)
    '100 m'
    >>> human_distance(km=0.4)
    '400 m'
    >>> human_distance(km=0.0366) # 40 Yard Dash
    '36.6 m'
    >>> human_distance(km=10.0000)
    '10 km'
    >>> human_distance(km=21.0975)
    '21.0975 km'
    >>> human_distance(km=42.1950)
    '42.195 km'
    """
    if km < 1:
        m = round(km * 1000, 1)
        if m == int(m):
            return "%i m" % m
        return "%.1f m" % m

    if km == int(km):
        return "%.0f km" % km

    # FIXME:
    txt = "%.4f" % km
    txt = txt.rstrip("0")
    return "%s km" % txt



def convert_cash_values(value, round_value=True):
    """
    >>> convert_cash_values(10.60)
    '11 €'
    >>> convert_cash_values(10.60, round_value=False)
    '10.60 €'
    """
    if round_value:
        return "%i %s" % (
            round(value), settings.FOR_RUNNERS_CURRENCY_SYMBOL
        )
    else:
        return "%.2f %s" % (
            value, settings.FOR_RUNNERS_CURRENCY_SYMBOL
        )




"""
    created 31.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import sys
from pathlib import Path

from django.contrib.auth.models import User
from django.core.management.base import BaseCommand
from django.db import IntegrityError
# https://github.com/jedie/django-for-runners
from for_runners.models import GpxModel


class Command(BaseCommand):
    help = "Import GPS files (*.gpx)"

    def add_arguments(self, parser):
        parser.add_argument(
            '--username',
            "-u",
            action='store',
            dest='username',
            required=True,
            help="The user to assign to the imported files"
        )
        parser.add_argument("path", help="Path to *.gpx files")

    def handle(self, *args, **options):
        username = options.get("username")
        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist as err:
            print("ERROR getting user %r: %s" % username, err)
            sys.exit(-1)

        path = Path(options.get("path"))
        path = path.expanduser()
        if not path.is_dir():
            print("ERROR: Given path '%s' is not a existing directory!" % path)
            sys.exit(-1)

        gpx_files = path.glob('**/*.gpx')
        for gpx_file in sorted(gpx_files):
            with gpx_file.open("r") as f:
                gpx_content = f.read()

            try:
                instance = GpxModel.objects.add_gpx(gpx_content, user)
            except IntegrityError as err:
                print("Skip .gpx file: %s" % err)
            else:
                if instance:
                    print("Add new track:", instance)



"""
    created 19.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import math
from datetime import datetime, timedelta

from gpxpy.gpx import GPX, GPXTrack, GPXTrackPoint, GPXTrackSegment

# circumference of the equator is 40075.017 km (WGS 84)
earth_circumference = 2 * math.pi * 6378.137  # ==> 40075.016685578485
km_per_longitude = earth_circumference / 360  # ==> 111.31949166666666 km


def lon2kilometers(lon_count):
    """
    return the distance in kilometers between the given longitude count.
    (Only valid at the equator: longitude=0 and elevation=0)

    >>> round(lon2kilometers(lon_count=1),2)  # 111.31949166666666 km
    111.32
    >>> round(lon2kilometers(lon_count=10),2) # 1113.1949166666666 km
    1113.19
    """
    return lon_count * km_per_longitude


def kilometers2lon_count(kilometers):
    """
    return the longitude count for the given kilometer distances.
    (Only valid at the equator: longitude=0 and elevation=0)

    >>> round(kilometers2lon_count(111.31949166666666),1)
    1.0
    >>> round(kilometers2lon_count(1113.1949166666666),1)
    10.0
    """
    return kilometers / km_per_longitude


def generate_gpx_track(
    track_length_km, point_count, pace_min, start_longitude=0, start_date=datetime(2018, 5, 30, 10, 00)
):
    distance_km = track_length_km / point_count
    print("km between points:", distance_km)

    longitude_diff = kilometers2lon_count(distance_km)
    print("longitude diff:", longitude_diff)

    time_delta = timedelta(minutes=(pace_min * distance_km))
    print("time delta:", time_delta)

    gpxpy_instance = GPX()
    gpxpy_instance.tracks.append(GPXTrack())
    gpxpy_instance.tracks[0].segments.append(GPXTrackSegment())
    points = gpxpy_instance.tracks[0].segments[0].points
    points.append(GPXTrackPoint(latitude=0, longitude=start_longitude, elevation=0, time=start_date))
    print("Start point:", points[-1])

    td = 0
    current_longitude = start_longitude
    current_datetime = start_date
    for point_no in range(point_count):
        current_longitude += longitude_diff
        current_datetime += time_delta
        points.append(GPXTrackPoint(latitude=0, longitude=current_longitude, elevation=0, time=current_datetime))
        print("point %i: %s" % (point_no + 1, points[-1]))

        print("\ttime diff:", points[-1].time_difference(points[-2]))
        print("\tdistance 2d:", points[-1].distance_2d(points[-2]))
        print("\tdistance 3d:", points[-1].distance_3d(points[-2]))
        td += points[-1].distance_3d(points[-2])
        print("\t", td)

    return gpxpy_instance



"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from django import forms


class UploadGpxFileForm(forms.Form):
    gpx_files = forms.FileField(widget=forms.ClearableFileInput(attrs={'multiple': True}))

    def send_email(self):
        # send email using the self.cleaned_data dictionary
        pass


INITIAL_DISTANCE = 2


class DistanceStatisticsForm(forms.Form):
    distance = forms.FloatField(
        label="Distance (km)",
        max_value=10,
        min_value=0.1,
        initial=INITIAL_DISTANCE,
        help_text="Gradation in kilometers to summarize the data."
    )






# Generated by Django 2.0.7 on 2018-07-12 20:43

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('for_runners', '0002_auto_20180710_1910'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='participationmodel',
            options={'ordering': ('-event__start_date', 'person'), 'verbose_name': 'Event Participation', 'verbose_name_plural': 'Event Participations'},
        ),
        migrations.RemoveField(
            model_name='gpxmodel',
            name='event',
        ),
        migrations.RemoveField(
            model_name='participationmodel',
            name='distance',
        ),
        migrations.AddField(
            model_name='gpxmodel',
            name='participation',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tracks', to='for_runners.ParticipationModel'),
        ),
        migrations.AddField(
            model_name='participationmodel',
            name='distance_km',
            field=models.DecimalField(decimal_places=4, default=0, help_text='Official track length in kilometer.', max_digits=7),
            preserve_default=False,
        ),
    ]



"""
    created 28.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from django.conf import settings
from django.core.management.base import BaseCommand
from for_runners.models import DistanceModel, GpxModel


class Command(BaseCommand):
    help = "Create some base data"

    def handle(self, *args, **options):

        updated_needed = False

        for distance_km in settings.BASE_IDEAL_TRACK_LENGTHS:
            obj, created = DistanceModel.objects.get_or_create(
                distance_km=distance_km)
            if created:
                updated_needed = True
                self.stdout.write("Create: %s" % repr(obj))
            else:
                self.stdout.write("%s already exists, ok." % repr(obj))

        if not updated_needed:
            print("No track updated needed, ok.")
        else:
            for track in GpxModel.objects.all():
                track.save()
                print(".", end="", flush=True)
            print()



import io
import logging

import svgwrite
from for_runners.exceptions import GpxDataError
from for_runners.gpx import get_2d_coordinate_list

log = logging.getLogger(__name__)


def gpx2svg(gpxpy_instance):
    """
    Optimize output with:

        * simplify(max_distance)
        * reduce_points(min_distance)

    FIXME:
        * Use correct World Geodetic System: WGS 84 calculations ;)
    """
    lat_list, lon_list = get_2d_coordinate_list(gpxpy_instance)

    lon_min = min(lon_list)
    lat_min = min(lat_list)
    lon_max = max(lon_list)
    lat_max = max(lat_list)

    log.debug("lon %s-%s lat %s-%s", lon_min, lon_max, lat_min, lat_max)

    lat_area = lat_max - lat_min
    lon_area = lon_max - lon_min
    if lon_area > lat_area:
        aspect = lat_area / lon_area
    else:
        aspect = lon_area / lat_area
    log.debug("Areas: %s,%s Aspect: %s", lat_area, lon_area, aspect)

    ############################################################################################

    border = 5

    # the minimum distance between two points
    min_distance = 0.2

    total_size_x = 100
    total_size_y = 100

    log.debug("total_size: %ix%i", total_size_x, total_size_y)

    drawing = svgwrite.Drawing(size=(total_size_x, total_size_y), profile='tiny')
    drawing.add(drawing.rect(insert=(0, 0), size=(total_size_x, total_size_y), fill='#000000'))
    drawing.add(drawing.rect(insert=(1, 1), size=(total_size_x - 2, total_size_y - 2), fill='#ffffff'))
    lines = drawing.add(drawing.g(stroke_width=1, stroke='blue', fill='none'))

    lines_size_x = total_size_x - (border * 2)
    lines_size_y = total_size_y - (border * 2)
    log.debug("lines_size: %ix%i", lines_size_x, lines_size_y)

    scale_x = lines_size_x / lon_area * aspect
    scale_y = lines_size_y / lat_area

    log.debug("scale: %ix%i", scale_x, scale_y)

    max_x = lon_area * scale_x
    max_y = lat_area * scale_y
    log.debug("max: %ix%i", max_x, max_y)

    offset_x = border + ((lines_size_x - max_x) / 2)
    offset_y = border + ((lines_size_y - max_y) / 2)
    log.debug("offset: %ix%i", offset_x, offset_y)

    # x_list = []
    # y_list = []

    old_x = None
    old_y = None
    for lon, lat in zip(lon_list, lat_list):
        x = ((lon - lon_min) * scale_x) + offset_x
        y = ((lat - lat_min) * scale_y) + offset_y

        y = y * -1 + total_size_y  # mirror the x-axis

        # x_list.append(x)
        # y_list.append(y)

        if old_x is not None:
            if abs(old_x - x) < min_distance:
                continue
            if abs(old_y - y) < min_distance:
                continue
            lines.add(drawing.line(start=(old_x, old_y), end=(x, y)))

        old_x = x
        old_y = y

    # log.debug(min(x_list), max(x_list), min(y_list), max(y_list))

    return drawing


def gpx2svg_file(gpxpy_instance, svg_filename, pretty=False):
    drawing = gpx2svg(gpxpy_instance)
    drawing.saveas(svg_filename, pretty=pretty)


def gpx2svg_string(gpxpy_instance, pretty=False):
    drawing = gpx2svg(gpxpy_instance)
    fileobj = io.StringIO()
    drawing.write(fileobj, pretty=pretty)
    return fileobj.getvalue().strip()


def gpx2svg_file(gpxpy_instance, fileobj, pretty=False):
    drawing = gpx2svg(gpxpy_instance)
    drawing.write(fileobj, pretty=pretty)



"""
    created 28.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
from for_runners.models.discipline import DisciplineModel
from for_runners.models.distance import DistanceModel
from for_runners.models.event import EventLinkModel, EventModel, ParticipationModel, CostModel
from for_runners.models.gpx import GpxModel



"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging
from urllib.parse import urlparse

from django.conf import settings
from django.db import models
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

# https://github.com/jedie/django-tools
from django_tools.models import UpdateInfoBaseModel, UpdateTimeBaseModel

# https://github.com/jedie/django-for-runners
from for_runners.gpx_tools.humanize import human_distance, human_seconds
from for_runners.model_utils import ModelAdminUrlMixin
from for_runners.models import DisciplineModel

log = logging.getLogger(__name__)


def human_url(url):
    scheme, netloc, url, params, query, fragment = urlparse(url)
    text = netloc + url
    text = text.strip("/")
    if text.startswith("www."):
        text = text[4:]
    return text


class LinkModelBase(UpdateTimeBaseModel):
    """
    inherit and automatically set from UpdateTimeBaseModel:
     * createtime
     * lastupdatetime
    """
    url = models.URLField(help_text=_("Link URL"))
    text = models.CharField(
        max_length=127,
        help_text=_("Link text (leave empty to generate it from url)"),
        null=True,
        blank=True,
    )
    title = models.CharField(
        max_length=127,
        help_text=_("Link title (leave empty to generate it from url)"),
        null=True,
        blank=True,
    )

    def get_text(self):
        return self.text or human_url(self.url)

    def get_title(self):
        return self.title or self.url

    def link_html(self):
        html = ('<a href="{url}" title="{title}" target="_blank">'
                '{text}'
                '</a>').format(
                    url=self.url, title=self.get_title(), text=self.get_text()
                )
        return mark_safe(html)

    link_html.short_description = _("Link")

    def save(self, *args, **kwargs):
        if self.text is None:
            self.text = human_url(self.url)

        super().save(*args, **kwargs)

    def __str__(self):
        return self.get_text()

    class Meta:
        abstract = True


class EventModel(ModelAdminUrlMixin, UpdateInfoBaseModel):
    """
    inherit from ModelAdminUrlMixin:
        * get_admin_change_url()

    inherit from UpdateInfoBaseModel:
        * createtime
        * lastupdatetime
        * createby
        * lastupdateby
    """
    no = models.PositiveIntegerField(
        help_text=_("Sequential number of the event"),
        null=True,
        blank=True,
    )
    name = models.CharField(max_length=255, help_text=_("Name of the event"))
    start_date = models.DateField(
        help_text=_("Start date of the run"),
        null=True,
        blank=True,
    )
    discipline = models.ForeignKey(DisciplineModel, on_delete=models.SET_NULL, null=True)

    def verbose_name(self):
        parts = []
        if self.no:
            parts.append("%i." % self.no)
        parts.append(self.name)
        if self.start_date:
            year = self.start_date.strftime("%Y")
            if year not in self.name:
                parts.append(year)
        result = " ".join([part for part in parts if part])
        return result

    verbose_name.short_description = _("Event Name")
    verbose_name.admin_order_field = "name"

    def links_html(self):
        links = []
        for link in self.links.all():
            links.append(link.link_html())
        html = "<br />".join(links)
        return mark_safe(html)

    links_html.short_description = _("Links")

    def __str__(self):
        return self.verbose_name()

    class Meta:
        verbose_name = _('Event')
        verbose_name_plural = _('Events')
        ordering = ('-start_date', '-pk')


class EventLinkModel(LinkModelBase):
    event = models.ForeignKey(EventModel, related_name="links", on_delete=models.CASCADE)


class ParticipationModel(ModelAdminUrlMixin, UpdateTimeBaseModel):
    """
    inherit from ModelAdminUrlMixin:
        * get_admin_change_url()

    inherit from UpdateTimeBaseModel:
        * createtime
        * lastupdatetime
    """
    event = models.ForeignKey(EventModel, related_name="participations", on_delete=models.CASCADE)
    person = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="+",
        help_text="The person who participated to this competition.",
        on_delete=models.CASCADE,
    )
    distance_km = models.DecimalField(
        help_text=_("Official track length in kilometer."),
        # store numbers up to 999 with a resolution of 4 decimal places
        max_digits=7,
        decimal_places=4,
    )
    duration = models.TimeField(
        verbose_name=_("Duration"),
        help_text=_("You officially measured finisher time"),
        null=True, blank=True,
    )
    start_number = models.CharField(max_length=15,
        help_text=_("Your start number"),
        null=True, blank=True
    )
    finisher_count = models.PositiveIntegerField(
        help_text=_("Number of participants who have finished in your discipline"),
        null=True, blank=True
    )

    def get_human_distance(self):
        return human_distance(self.distance_km)

    get_human_distance.short_description = _("Distance")
    get_human_distance.admin_order_field = "distance_km"

    def get_duration_s(self):
        """
        :return: duration in seconds
        """
        if self.duration:
            # FIXME: Is there really no easier way to do this?
            duration = self.duration.second
            duration += (self.duration.minute * 60)
            duration += (self.duration.hour * 60 * 60)
            return duration

    def human_duration(self):
        if self.duration:
            return human_seconds(self.get_duration_s())

    def get_pace_s(self):
        if self.duration:
            pace = (self.get_duration_s() / 60) / float(self.distance_km)
            if pace > 99 or pace < 0:
                log.error("Pace out of range: %f", pace)
            else:
                return pace * 60

    def verbose_name(self):
        parts = [
            self.event.verbose_name(),
            "-",
            self.person.username,
            "-",
            self.get_human_distance(),
        ]
        if self.duration:
            parts.append("in %s" % self.human_duration())

        result = " ".join([part for part in parts if part])
        return result

    verbose_name.short_description = _("Event Name")
    verbose_name.admin_order_field = "name"

    def __str__(self):
        return self.verbose_name()

    class Meta:
        verbose_name = _('Event Participation')
        verbose_name_plural = _('Event Participations')
        ordering = ('-event__start_date', 'person')


class CostModel(UpdateTimeBaseModel):
    """
    inherit and automatically set from UpdateTimeBaseModel:
     * createtime
     * lastupdatetime
    """
    participation = models.ForeignKey(ParticipationModel, related_name="costs", on_delete=models.CASCADE)
    name = models.CharField(max_length=15,
        help_text=_("The name of the item you dis pay for"),
    )
    amount = models.DecimalField(
        help_text=_("How much did you pay for this?"),
        max_digits=8, decimal_places=2,  # store numbers up to 99 with a resolution of 2 decimal places
    )

    def __str__(self):
        return "%s: %s" % (self.name, self.amount)

    class Meta:
        verbose_name = _('Participation Cost')
        verbose_name_plural = _('Participation Costs')



"""
    created 19.07.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
from django.views import generic
from django.views.generic import TemplateView


class ChangelistViewMixin:

    def dispatch(self, request, change_list, *args, **kwargs):
        self.change_list = change_list
        return super().dispatch(request, *args, **kwargs)


class BaseChangelistView(ChangelistViewMixin, TemplateView):
    """
    Baseclass for chnagelist views without forms.
    """
    pass


class BaseFormChangelistView(ChangelistViewMixin, generic.FormView):
    """
    Baseclass for changelist views with forms.
    """
    form_class = None

    def form_valid(self, form):
        # Don't redirect, if form is valid ;)
        return self.render_to_response(self.get_context_data(form=form))



"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

__version__ = '0.7.1'



# coding: utf-8

from django.conf.urls import include, static, url
from django.conf.urls.i18n import i18n_patterns
from django.contrib import admin
from django.views.generic import RedirectView
from django.urls import path
# https://github.com/jedie/django-for-runners
from for_runners_test_project import settings

admin.autodiscover()

urlpatterns = i18n_patterns(
    path('admin/', admin.site.urls),

    # until there is not real CMS pages: redirect to the interesting admin page:
    url(r"^$", RedirectView.as_view(url="/admin/for_runners/gpxmodel/")),
)

if settings.DEBUG:
    urlpatterns += static.static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)
    urlpatterns += static.static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

    import debug_toolbar
    urlpatterns = [
        url(r'^__debug__/', include(debug_toolbar.urls)),
    ] + urlpatterns



# coding: utf-8
from pathlib import Path

import pytest
# https://github.com/jedie/django-tools
from django.core.files.uploadedfile import UploadedFile
from django_tools.unittest_utils.BrowserDebug import debug_response
from django_tools.unittest_utils.unittest_base import BaseTestCase
from django_tools.unittest_utils.user import TestUserMixin
from for_runners.models import GpxModel
from for_runners.version import __version__

BASE_PATH = Path(__file__).parent


class ForRunnerAdminTests(TestUserMixin, BaseTestCase):
    """
        Special for runners admin tests
    """

    def test_staff_upload(self):
        self.login(usertype='superuser')

        gpx_file_path1 = Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx")
        gpx_file_path2 = Path(BASE_PATH, "fixture_files/no_track_points.gpx")

        with gpx_file_path1.open("rb") as file1, gpx_file_path2.open("rb") as file2:

            response = self.client.post(
                '/en/admin/for_runners/gpxmodel/upload/',
                data={
                    "gpx_files": [
                        UploadedFile(file=file1, name=gpx_file_path1.name, content_type="application/gpx+xml"),
                        UploadedFile(file=file2, name=gpx_file_path2.name, content_type="application/gpx+xml")
                    ]
                },
                HTTP_ACCEPT_LANGUAGE='en'
            )
            # debug_response(response)

        tracks = GpxModel.objects.all()
        self.assertEqual(tracks.count(), 1)
        new_track = tracks[0]

        self.assertRedirects(
            response,
            expected_url="/en/admin/for_runners/gpxmodel/%i/change/" % new_track.pk,
            status_code=302,
            target_status_code=200,
            fetch_redirect_response=False
        )

        response = self.client.get("/en/admin/for_runners/gpxmodel/", HTTP_ACCEPT_LANGUAGE='en')
        # print(repr(self.get_messages(response)))
        self.assertResponse(
            response,
            must_contain=(
                '<title>Select GPX Track to change | Django-ForRunners v%s</title>' % __version__,

                # "Process garmin_connect_1.gpx...",
                # "Created: 2018-02-21 14:30:50 Moers Hülsdonk",
                #
                # "Process no_track_points.gpx...",
                # "Error process GPX data: Can't get first track",
                '<td class="field-human_pace">7:03 min/km</td>',
                '<p class="paginator">1 GPX Track</p>',
            ),
            messages=[
                'Process garmin_connect_1.gpx...', 'Created: 2018-02-21 Moers Hülsdonk',
                'Process no_track_points.gpx...', "Error process GPX data: Can't get first track"
            ],
            must_not_contain=('error', 'traceback'),
            template_name='admin/for_runners/gpxmodel/change_list.html',
            html=True,
        )

    def test_staff_add_view(self):
        self.login(usertype='superuser')
        response = self.client.get("/en/admin/for_runners/gpxmodel/add/", HTTP_ACCEPT_LANGUAGE='en')
        self.assertResponse(
            response,
            must_contain=(
                '<title>Add GPX Track | Django-ForRunners v%s</title>' % __version__,
                '<textarea name="gpx" cols="40" rows="10" class="vLargeTextField" required id="id_gpx"></textarea>',
            ),
            messages=[],
            must_not_contain=('error', 'traceback'),
            template_name='admin/for_runners/gpxmodel/change_form.html',
            html=True,
        )

    def test_staff_add_gpx(self):

        # GpxModel.objects.all().delete()

        gpx_file_path1 = Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx")
        with gpx_file_path1.open("rb") as f:
            gpx_content = f.read().decode("utf-8")

        self.login(usertype='superuser')
        response = self.client.post(
            path="http://127.0.0.1:8000/en/admin/for_runners/gpxmodel/add/",
            data={
                "gpx": gpx_content,
            },
            HTTP_ACCEPT_LANGUAGE="en"
        )
        self.assertRedirects(
            response,
            expected_url="/en/admin/for_runners/gpxmodel/",
            status_code=302,
            target_status_code=200,
            fetch_redirect_response=False
        )

        response = self.client.get("/en/admin/for_runners/gpxmodel/1/change/", HTTP_ACCEPT_LANGUAGE='en')
        self.assertResponse(
            response,
            must_contain=(
                '<title>Change GPX Track | Django-ForRunners v%s</title>' % __version__,
                '<label>Start time:</label>',
                '<div class="readonly">Feb. 21, 2018, 2:30 p.m.</div>',
                '<input type="text" name="short_start_address" value="Moers Hülsdonk" style="width:70%" maxlength="255" id="id_short_start_address" />',
                '<label>Finish time:</label>',
                '<div class="readonly">Feb. 21, 2018, 2:30 p.m.</div>',
                '<input type="text" name="short_finish_address" value="Moers Hülsdonk" style="width:70%" maxlength="255" id="id_short_finish_address" />',
            ),
            messages=[
                'The GPX Track'
                ' "<a href="/en/admin/for_runners/gpxmodel/1/change/">2018-02-21 Moers Hülsdonk</a>"'
                ' was added successfully.'
            ],
            must_not_contain=('error', 'traceback'),
            template_name='admin/for_runners/gpxmodel/change_form.html',
            html=True,
        )



from __future__ import absolute_import

from .version import __version__  # noqa

# https://docs.djangoproject.com/en/2.0/ref/applications/#configuring-applications-ref
default_app_config = "for_runners.apps.ForRunnersConfig"



"""
    created 2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import math
import operator
import unittest
from datetime import datetime
from functools import reduce
from pathlib import Path
# https://github.com/jedie/django-for-runners
from pprint import pprint

from for_runners.gpx import (GpxIdentifier, GpxMedian, iter_distances, parse_gpx_file)
from for_runners.tests.base import BaseTestCase
from for_runners.tests.utils import (earth_circumference, generate_gpx_track, kilometers2lon_count, lon2kilometers)
from gpxpy.gpx import GPX, GPXTrack, GPXTrackPoint, GPXTrackSegment

BASE_PATH = Path(__file__).parent

class GpxTests(BaseTestCase):

    def test_parse_gpx(self):
        filepath = Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx")

        gpxpy_instance = parse_gpx_file(filepath)

        first_point = gpxpy_instance.tracks[0].segments[0].points[0]
        self.assertEqual(first_point.latitude, 51.43788929097354412078857421875)
        self.assertEqual(first_point.longitude, 6.617012657225131988525390625)

        extensions = first_point.extensions
        self.assertEqual(1, len(extensions))
        self.assertEqual(extensions[0].getchildren()[0].text.strip(), "120")  # hr
        self.assertEqual(extensions[0].getchildren()[1].text.strip(), "70")  # cad

    def test_parse_noi_track_gpx(self):
        filepath = Path(BASE_PATH, "fixture_files/no_track_points.gpx")

        gpxpy_instance = parse_gpx_file(filepath)
        print(gpxpy_instance)
        self.assertEqual(
            repr(gpxpy_instance),
            "GPX(waypoints=[GPXWaypoint(51.43788, 6.61701, name='Somewhere', description='foobar', symbol='Waypoint')])"
        )

    def test_parse_no_heart_rate_gpx(self):
        filepath = Path(BASE_PATH, "fixture_files/garmin_no_heart_rate.gpx")

        gpxpy_instance = parse_gpx_file(filepath)
        print(gpxpy_instance)
        self.assertEqual(
            repr(gpxpy_instance), "GPX(tracks=[GPXTrack(name='Foo Bar', segments=[GPXTrackSegment(points=[...])])])"
        )

    def test_iter_distance(self):

        # containes 3 points:
        filepath = Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx")

        gpxpy_instance = parse_gpx_file(filepath)
        distances = []
        total_points = []
        total_distance = 0
        for section_distance, points in iter_distances(gpxpy_instance, distance=1):
            distances.append(section_distance)
            total_distance += section_distance
            for point in points:
                total_points.append((point.latitude, point.longitude))

        self.assertEqual(
            total_points, [(51.437889290973544, 6.617012657225132), (51.43786800093949, 6.617006119340658),
                           (51.437847297638655, 6.6170057002455)]
        )

        self.assert_equal_rounded(gpxpy_instance.length_3d(), 4.726553499192461)
        self.assert_equal_rounded(total_distance, 4.726553499192461)

        self.assertEqual(distances, [0, 2.413028183109784, 2.313525316082677])


class GpxMedianTests(BaseTestCase):

    def test_garmin_connect_1_gpx(self):
        # containes 3 points:
        filepath = Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx")
        total_distance = 4.726553499192461

        gpxpy_instance = parse_gpx_file(filepath)

        gpx_median = GpxMedian(gpxpy_instance, distance=1)
        self.assertEqual(gpx_median.total_distance, total_distance)

        distances = tuple(gpx_median.iter_section_distance())
        self.assertEqual(reduce(operator.add, distances), total_distance)
        self.assertEqual(distances, (0, 2.413028183109784, 2.313525316082677))

    def test_iter_extension_data(self):
        gpx_median = GpxMedian(
            gpxpy_instance = parse_gpx_file(Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx")),
            distance=1
        )

        hr_data = tuple(gpx_median.iter_extension_data(key="hr"))
        pprint(hr_data)
        self.assertEqual(hr_data, (([120], 120, 120, 120), ([125], 125, 125, 125), ([130], 130, 130, 130)))

        cad_data = tuple(gpx_median.iter_extension_data(key="cad"))
        pprint(cad_data)
        self.assertEqual(cad_data, (([70], 70, 70, 70), ([75], 75, 75, 75), ([80], 80, 80, 80)))

    def test_latitude_calculation(self):
        gpxpy_instance = GPX()
        gpxpy_instance.tracks.append(GPXTrack())
        gpxpy_instance.tracks[0].segments.append(GPXTrackSegment())
        points = gpxpy_instance.tracks[0].segments[0].points

        longitude_distance_km = 111.31949079327357

        self.assertEqual(lon2kilometers(lon_count=1), longitude_distance_km)
        self.assertEqual(kilometers2lon_count(longitude_distance_km), 1)

        for lon_count in range(10):
            distance_km = lon2kilometers(lon_count)
            self.assert_equal_rounded(distance_km, longitude_distance_km*lon_count)
            self.assert_equal_rounded(kilometers2lon_count(distance_km), lon_count, decimal_places=0)

    def test_generate_gpx_track(self):

        track_length_km=earth_circumference/2

        gpxpy_instance = generate_gpx_track(track_length_km=track_length_km, point_count=180, pace_min=6, start_longitude=0, start_date=datetime(2018, 5, 30, 10, 00))
        print(gpxpy_instance.to_xml())

        self.assertEqual(gpxpy_instance.get_points_no(), 181) # 180 + No.0

        gpxpy_length_2d_km = gpxpy_instance.length_2d()/1000
        self.assert_equal_rounded(gpxpy_length_2d_km, track_length_km)

        gpxpy_length_3d_km = gpxpy_instance.length_3d()/1000
        self.assert_equal_rounded(gpxpy_length_3d_km, track_length_km)

        # 20015.08679602057
        # 20015.086796020572


class GpxIdentifierTests(BaseTestCase):

    def setUp(self):
        gpxpy_instance = parse_gpx_file(Path(BASE_PATH, "fixture_files/garmin_connect_1.gpx"))
        self.gi = GpxIdentifier(gpxpy_instance)

    def test_identifier_string(self):
        self.assertEqual(
            self.gi._identifier_string(), (
                "20180221143050"
                "_51.4378892909735441207885742"
                "_6.6170126572251319885253906"
                "_20180221143052"
                "_51.4378472976386547088623047"
                "_6.6170057002454996109008789"
            )
        )

    def test_prefix(self):
        self.assertEqual(self.gi._prefix(), "20180221_1430")

    def test_prefix_id(self):
        self.assertEqual(self.gi.get_prefix_id(), "20180221_1430_a307a8")



"""
    Convert a Garmin connect GPX file to a gpxpy instance

    work-a-round until https://github.com/tkrajina/gpxpy/issues/115#issuecomment-392798245 fixed

    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from pathlib import Path

import gpxpy.gpx
from lxml import etree


def garmin2gpxpy(content):
    gpx = gpxpy.gpx.GPX()
    gpx.tracks.append(gpxpy.gpx.GPXTrack())
    gpx.tracks[0].segments.append(gpxpy.gpx.GPXTrackSegment())

    NS = {'ns': 'http://www.topografix.com/GPX/1/1'}

    if isinstance(content, str):
        content = content.encode("utf-8")

    tree = etree.fromstring(content)

    for track_point in tree.xpath('//ns:trk/ns:trkseg/ns:trkpt', namespaces=NS):
        latitude = float(track_point.get("lat"))
        longitude = float(track_point.get("lon"))

        elevation = track_point.xpath('ns:ele/text()', namespaces=NS)
        elevation = float(elevation[0])

        point_time = track_point.xpath('ns:time/text()', namespaces=NS)[0]
        # e.g.: 2018-04-28T14:30:50.000Z
        point_time = gpxpy.gpxfield.parse_time(point_time)

        # print('latitude', latitude, 'longitude', longitude, "elevation", elevation)
        point = gpxpy.gpx.GPXTrackPoint(latitude=latitude, longitude=longitude, elevation=elevation, time=point_time)
        gpx.tracks[0].segments[0].points.append(point)

    return gpx




"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import logging

from django.db import models

from gpxpy.gpx import GPXException

# https://github.com/jedie/django-for-runners
from for_runners.gpx import get_identifier, parse_gpx

log = logging.getLogger(__name__)


class GpxModelQuerySet(models.QuerySet):

    def get_by_identifier(self, identifier):
        """
        :param identifier: 'Identifier' namedtuple created here: for_runners.gpx.get_identifier
        """
        return self.get(
            start_time=identifier.start_time,
            finish_time=identifier.finish_time,
            start_latitude=identifier.start_lat,
            start_longitude=identifier.start_lon,
            finish_latitude=identifier.finish_lat,
            finish_longitude=identifier.finish_lon,
        )


class BaseGpxModelManager(models.Manager):

    def get_queryset(self):
        return GpxModelQuerySet(self.model, using=self._db)

    def add_gpx(self, gpx_content, user):
        """
        Create a new for_runners.models.GpxModel entry

        :param gpx_content: String content of the new gpx file
        :return: GpxModel instance
        """
        try:
            gpxpy_instance = parse_gpx(gpx_content)
        except GPXException as err:
            log.exception("Invalid GPX Data: %s" % err)
            return

        identifier = get_identifier(gpxpy_instance)

        qs = self.get_queryset()

        try:
            instance = qs.get_by_identifier(identifier)
        except self.model.DoesNotExist:
            log.debug("Create new track for user: %s", user)
            instance = self.create(
                gpx=gpx_content,
                tracked_by=user,
            )
            return instance
        else:
            log.info("Skip existing track: %s", instance)
            return


GpxModelManager = BaseGpxModelManager.from_queryset(GpxModelQuerySet)



"""
    created 31.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import collections
import hashlib
import io
import statistics
from pathlib import Path

import gpxpy
# https://github.com/jedie/django-for-runners
from for_runners.exceptions import GpxDataError
from for_runners.gpx_tools.garmin2gpxpy import garmin2gpxpy
from gpxpy.geo import distance as geo_distance

Identifier = collections.namedtuple(
    'Identifier', ('start_time, start_lat, start_lon, finish_time, finish_lat, finish_lon')
)


def get_identifier(gpxpy_instance):
    """
    :return: Identifier named tuple
    """
    time_bounds = gpxpy_instance.get_time_bounds()

    try:
        first_track = gpxpy_instance.tracks[0]
    except IndexError:
        raise GpxDataError("Can't get first track")

    try:
        first_segment = first_track.segments[0]
    except IndexError:
        raise GpxDataError("Can't get first segment")

    try:
        first_point = first_segment.points[0]
    except IndexError:
        raise GpxDataError("Can't get first segment point")

    try:
        last_track = gpxpy_instance.tracks[-1]
    except IndexError:
        raise GpxDataError("Can't get last track")

    try:
        last_segment = last_track.segments[-1]
    except IndexError:
        raise GpxDataError("Can't get last segment")

    try:
        last_point = last_segment.points[-1]
    except IndexError:
        raise GpxDataError("Can't get last segment point")

    return Identifier(
        time_bounds.start_time, first_point.latitude, first_point.longitude, time_bounds.end_time, last_point.latitude,
        last_point.longitude
    )


class GpxIdentifier:
    HASHER_NAME = "sha512"
    HASH_LENGTH = 6

    def __init__(self, gpxpy_instance):
        self.identifier = get_identifier(gpxpy_instance)
        self.hasher = hashlib.new(self.HASHER_NAME)

    def _identifier_string(self):
        data = []
        data.append(self.identifier.start_time.strftime("%Y%m%d%H%M%S"))
        data.append("%.25f" % self.identifier.start_lat)
        data.append("%.25f" % self.identifier.start_lon)
        data.append(self.identifier.finish_time.strftime("%Y%m%d%H%M%S"))
        data.append("%.25f" % self.identifier.finish_lat)
        data.append("%.25f" % self.identifier.finish_lon)
        return "_".join(data)

    def _prefix(self):
        return self.identifier.start_time.strftime("%Y%m%d_%H%M")

    def get_prefix_id(self):
        identifier_string = self._identifier_string()
        self.hasher.update(identifier_string.encode("ASCII"))
        hex_digest = self.hasher.hexdigest()
        hex_digest_cut = hex_digest[:self.HASH_LENGTH]
        prefix = self._prefix()
        result = "%s_%s" % (prefix, hex_digest_cut)
        return result


def parse_gpx(content):
    # if 'creator="Garmin Connect"' in content:
    #     # work-a-round until https://github.com/tkrajina/gpxpy/issues/115#issuecomment-392798245 fixed
    #     return garmin2gpxpy(content)

    return gpxpy.parse(content)


def parse_gpx_file(filepath):
    assert filepath.is_file(), "File not found: '%s'" % filepath
    with filepath.open("r") as f:
        content = f.read()

    return parse_gpx(content)


def iter_points(gpxpy_instance):
    for track in gpxpy_instance.tracks:
        for segment in track.segments:
            for point in segment.points:
                yield point


def iter_coordinates(gpxpy_instance):
    for point in iter_points(gpxpy_instance):
        # log.debug('Point at ({0},{1}) -> {2}'.format(point.latitude, point.longitude, point.elevation))
        yield (point.latitude, point.longitude, point.elevation)


def get_2d_coordinate_list(gpxpy_instance):
    lat_list = []
    lon_list = []
    for latitude, longitude, elevation in iter_coordinates(gpxpy_instance):
        lat_list.append(latitude)
        lon_list.append(longitude)

    if not lat_list or not lon_list:
        raise GpxDataError("No track points in file!")

    return (lat_list, lon_list)


def get_2d_coordinate_list(gpxpy_instance):
    lat_list = []
    lon_list = []
    for latitude, longitude, elevation in iter_coordinates(gpxpy_instance):
        lat_list.append(latitude)
        lon_list.append(longitude)

    if not lat_list or not lon_list:
        raise GpxDataError("No track points in file!")

    return (lat_list, lon_list)


def add_extension_data(point):
    """
    Add all existing extension_data dict to the point.
    Garmin can contain:
        'hr' -> heart rate
        'cad' -> cadence value
    """
    extension_data = {}
    extensions = point.extensions
    if extensions:
        for child in extensions[0].getchildren():
            tag = child.tag.rsplit("}", 1)[-1]  # FIXME

            value = child.text
            try:
                if "." in value:
                    value = float(value)
                else:
                    value = int(value)
            except ValueError:
                pass
            extension_data[tag] = value

    point.extension_data = extension_data
    return point


def iter_extensions(points):
    for point in points:
        point = add_extension_data(point)  # set: point.extension_data dict
        yield point


def get_extension_data(gpxpy_instance):
    """
    return a dict with all extension values from all track points.
    """
    extension_data = collections.defaultdict(list)
    for point in iter_points(gpxpy_instance):
        point = add_extension_data(point)  # set: point.extension_data dict
        for key, value in point.extension_data.items():
            extension_data[key].append(value)

    return dict(extension_data)


def iter_distance(gpxpy_instance, distance):
    """
    iterate over GPXTrackPoint() instances at intervals of <distance> in meters
    """

    def convert_distance_in_km(distance):
        """
        Calculate the rounded kilometers as integer from the current distance in meters.
        """
        return int(round(distance / 1000))

    iterator = iter_points(gpxpy_instance)

    previous_point = next(iterator)
    old_latitude, old_longitude, old_elevation = previous_point.latitude, previous_point.longitude, previous_point.elevation

    count = 1
    next_distance = distance
    total_distance = 0
    for point in iterator:
        latitude, longitude, elevation = point.latitude, point.longitude, point.elevation

        previous_total_distance = total_distance
        total_distance += geo_distance(old_latitude, old_longitude, old_elevation, latitude, longitude, elevation)

        old_latitude, old_longitude, old_elevation = latitude, longitude, elevation

        if total_distance >= next_distance:

            # Calculate the deviation from the ideal distance:
            current_difference = total_distance - next_distance
            previous_difference = abs(previous_total_distance - next_distance)

            print(
                "no. %03i: previous point %.1fm diff: %.1fm vs. current point %.1fm diff: %.1fm" %
                (count, previous_total_distance, previous_difference, total_distance, current_difference)
            )

            # We didn't use the >count< for resulting kilometers:
            # Maybe the point density is very low and the real distance is greater than kilometers ;)
            # Use convert_distance_in_km() and the real distance.

            if previous_difference < current_difference:
                # The deviation from the previous position is smaller:
                yield previous_point, previous_total_distance, convert_distance_in_km(previous_total_distance)
            else:
                yield point, total_distance, convert_distance_in_km(total_distance)

            count += 1
            next_distance = distance * count

        previous_point = point


def iter_distances(gpxpy_instance, distance):
    """
    iterate over GPXTrackPoint() instances at intervals of <distance> in meters
    """
    iterator = iter_points(gpxpy_instance)

    previous_point = next(iterator)
    points = [previous_point]
    old_latitude, old_longitude, old_elevation = previous_point.latitude, previous_point.longitude, previous_point.elevation

    count = 1
    next_distance = distance
    section_distance = 0
    for point in iterator:
        point = add_extension_data(point)  # set: point.extension_data dict
        points.append(point)
        latitude, longitude, elevation = point.latitude, point.longitude, point.elevation

        point_distance = geo_distance(old_latitude, old_longitude, old_elevation, latitude, longitude, elevation)

        previous_section_distance = section_distance
        section_distance += point_distance

        old_latitude, old_longitude, old_elevation = latitude, longitude, elevation

        if section_distance >= next_distance:

            # Calculate the deviation from the ideal distance:
            current_difference = section_distance - next_distance
            previous_difference = abs(previous_section_distance - next_distance)

            print(
                "no. %03i: previous point %.1fm diff: %.1fm vs. current point %.1fm diff: %.1fm" %
                (count, previous_section_distance, previous_difference, section_distance, current_difference)
            )

            # We didn't use the >count< for resulting kilometers:
            # Maybe the point density is very low and the real distance is greater than kilometers ;)
            # Use convert_distance_in_km() and the real distance.

            if previous_difference < current_difference:
                # The deviation from the previous position is smaller:
                yield previous_section_distance, points[:-1]
                points = [point]
                section_distance = point_distance
            else:
                yield section_distance, points
                points = []
                section_distance = 0

            count += 1
            next_distance = distance * count

    # the last points
    yield section_distance, points


class GpxSection:

    def __init__(self, distance, points):
        self.distance = distance
        self.points = points

        self.extension_data = collections.defaultdict(list)
        for point in points:
            point = add_extension_data(point)  # set: point.extension_data dict
            for key, value in point.extension_data.items():
                self.extension_data[key].append(value)

        self.extension_data_median = {}

        for key in self.extension_data.keys():
            values = self.extension_data[key]
            print(key, values)

            self.extension_data_median["%s_min" % key] = min(values)
            self.extension_data_median["%s_max" % key] = max(values)
            self.extension_data_median["%s_avg" % key] = statistics.median(values)

    def get_extension_min(self, key):
        return self.extension_data_median["%s_min" % key]

    def get_extension_max(self, key):
        return self.extension_data_median["%s_max" % key]

    def get_extension_avg(self, key):
        return self.extension_data_median["%s_avg" % key]


class GpxMedian:

    def __init__(self, gpxpy_instance, distance):
        self.total_distance = 0
        self.sections = []
        for section_distance, points in iter_distances(gpxpy_instance, distance=distance):
            self.total_distance += section_distance
            self.sections.append(GpxSection(section_distance, points))

    def iter_section_distance(self):
        for section in self.sections:
            yield section.distance

    def iter_extension_data(self, key):
        for section in self.sections:
            min_value = section.get_extension_min(key)
            max_value = section.get_extension_max(key)
            avg_value = section.get_extension_avg(key)
            yield section.extension_data[key], min_value, max_value, avg_value



"""
    created 19.07.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
from for_runners.admin.discipline import DisciplineModelAdmin
from for_runners.admin.distance import DistanceModelAdmin
from for_runners.admin.event import EventLinkModelAdmin, EventModelAdmin, ParticipationModelAdmin
from for_runners.admin.gpx import GpxModelAdmin



"""
    TODO: Move to django-tools
"""

from django.urls import reverse


class ModelAdminUrlMixin:
    def get_admin_change_url(self):
        return reverse(
            'admin:{0}_{1}_change'.format(
                self._meta.app_label,
                self._meta.model_name,
            ),
            args=(self.pk,)
        )



# Generated by Django 2.0.7 on 2018-07-04 23:08

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import for_runners.models.gpx


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='DisciplineModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Sport discipline', max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='DistanceModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('distance_km', models.DecimalField(decimal_places=4, help_text='The ideal track length in kilometer.', max_digits=7, unique=True)),
                ('variance', models.PositiveSmallIntegerField(default=5, help_text='Maximum (+/-) deviation in percent to match this distance.')),
                ('min_distance_m', models.PositiveIntegerField(blank=True, editable=False, null=True)),
                ('max_distance_m', models.PositiveIntegerField(blank=True, editable=False, null=True)),
            ],
            options={
                'verbose_name': 'Distance',
                'verbose_name_plural': 'Distances',
                'ordering': ('-distance_km',),
            },
        ),
        migrations.CreateModel(
            name='EventLinkModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('createtime', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='Create time')),
                ('lastupdatetime', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='Time of the last change.')),
                ('url', models.URLField(help_text='Link URL')),
                ('text', models.CharField(blank=True, help_text='Link text (leave empty to generate it from url)', max_length=127, null=True)),
                ('title', models.CharField(blank=True, help_text='Link title (leave empty to generate it from url)', max_length=127, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='EventModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('createtime', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='Create time')),
                ('lastupdatetime', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='Time of the last change.')),
                ('no', models.PositiveIntegerField(blank=True, help_text='Sequential number of the event', null=True)),
                ('name', models.CharField(help_text='Name of the event', max_length=255)),
                ('start_date', models.DateField(blank=True, help_text='Start date of the run', null=True)),
                ('createby', models.ForeignKey(blank=True, editable=False, help_text='User how create this entry.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='eventmodel_createby', to=settings.AUTH_USER_MODEL)),
                ('discipline', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='for_runners.DisciplineModel')),
                ('lastupdateby', models.ForeignKey(blank=True, editable=False, help_text='User as last edit this entry.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='eventmodel_lastupdateby', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Event',
                'verbose_name_plural': 'Events',
                'ordering': ('-start_date', '-pk'),
            },
        ),
        migrations.CreateModel(
            name='GpxModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('createtime', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='Create time')),
                ('lastupdatetime', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='Time of the last change.')),
                ('gpx', models.TextField(help_text='The raw gpx file content')),
                ('creator', models.CharField(blank=True, help_text='Used device to create this track', max_length=511, null=True)),
                ('track_svg', models.FileField(blank=True, null=True, upload_to=for_runners.models.gpx.svg_upload_path, verbose_name='Track SVG')),
                ('start_time', models.DateTimeField(editable=False, help_text='Start time of the first segment in track')),
                ('start_latitude', models.FloatField(editable=False, help_text='Latitude of the first recorded point from the *.gpx file')),
                ('start_longitude', models.FloatField(editable=False, help_text='Longitude of the first recorded point from the *.gpx file')),
                ('start_temperature', models.FloatField(blank=True, help_text='Temperature at start.', null=True)),
                ('start_weather_state', models.CharField(blank=True, help_text='Weather state at start.', max_length=127, null=True)),
                ('short_start_address', models.CharField(blank=True, help_text='The short address of the start point', max_length=255, null=True)),
                ('full_start_address', models.CharField(blank=True, help_text='The full address of the start point', max_length=255, null=True)),
                ('finish_time', models.DateTimeField(editable=False, help_text='End time of the last segment in track')),
                ('finish_latitude', models.FloatField(editable=False, help_text='Latitude of the finish point')),
                ('finish_longitude', models.FloatField(editable=False, help_text='Longitude of the finish point')),
                ('finish_temperature', models.FloatField(blank=True, help_text='Temperature at finish.', null=True)),
                ('finish_weather_state', models.CharField(blank=True, help_text='Weather state at finish.', max_length=127, null=True)),
                ('short_finish_address', models.CharField(blank=True, help_text='The short address of the finish point', max_length=255, null=True)),
                ('full_finish_address', models.CharField(blank=True, help_text='The full address of the finish point', max_length=255, null=True)),
                ('points_no', models.PositiveIntegerField(blank=True, help_text='Number of points in GPX', null=True)),
                ('length', models.PositiveIntegerField(blank=True, help_text='Length in meters (calculated 3-dimensional used latitude, longitude, and elevation)', null=True)),
                ('duration', models.PositiveIntegerField(blank=True, help_text='Duration in seconds', null=True)),
                ('net_duration', models.TimeField(blank=True, help_text='The officially measured time (e.g. from the official timekeeping of a running event.)', null=True, verbose_name='Net Duration')),
                ('pace', models.DecimalField(blank=True, decimal_places=2, help_text='Min/km (number of minutes it takes to cover a kilometer)', max_digits=4, null=True)),
                ('uphill', models.IntegerField(blank=True, help_text='Uphill elevation climbs in meters', null=True)),
                ('downhill', models.IntegerField(blank=True, help_text='Downhill elevation descent in meters', null=True)),
                ('min_elevation', models.IntegerField(blank=True, help_text='Minimum elevation in meters', null=True)),
                ('max_elevation', models.IntegerField(blank=True, help_text='Maximum elevation in meters', null=True)),
                ('heart_rate_min', models.PositiveIntegerField(blank=True, editable=False, help_text='Minimum heart rate.', null=True)),
                ('heart_rate_avg', models.PositiveIntegerField(blank=True, editable=False, help_text='Average heart rate.', null=True)),
                ('heart_rate_max', models.PositiveIntegerField(blank=True, editable=False, help_text='Maximum heart rate.', null=True)),
                ('event', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tracks', to='for_runners.EventModel')),
                ('ideal_distance', models.ForeignKey(blank=True, help_text='Length in meters (calculated 3-dimensional used latitude, longitude, and elevation)', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tracks', to='for_runners.DistanceModel')),
                ('lastupdateby', models.ForeignKey(blank=True, editable=False, help_text='User as last edit this entry', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='gpxmodel_lastupdateby', to=settings.AUTH_USER_MODEL)),
                ('tracked_by', models.ForeignKey(blank=True, editable=False, help_text='The user that tracked this gpx entry', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='gpxmodel_createby', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'GPX Track',
                'verbose_name_plural': 'GPX Tracks',
                'ordering': ('-start_time', '-pk'),
            },
        ),
        migrations.AddField(
            model_name='eventlinkmodel',
            name='event',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='links', to='for_runners.EventModel'),
        ),
        migrations.AlterUniqueTogether(
            name='gpxmodel',
            unique_together={('start_time', 'start_latitude', 'start_longitude', 'finish_time', 'finish_latitude', 'finish_longitude')},
        ),
    ]



# coding: utf-8

"""
    Django settings for test_project project.

    Good startpoint fot Django-CMS stuff is:

    https://github.com/nephila/djangocms-installer/blob/develop/djangocms_installer/config/settings.py


    For more information on this file, see
    https://docs.djangoproject.com/en/1.8/topics/settings/

    For the full list of settings and their values, see
    https://docs.djangoproject.com/en/1.8/ref/settings/
"""


print("Use settings:", __file__)


import os

from django.utils.translation import ugettext_lazy as _


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.abspath(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.8/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'Only for the tests ;)'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

SITE_ID=1


from django_tools.settings_utils import FnMatchIps

# Required for the debug toolbar to be displayed:
INTERNAL_IPS = FnMatchIps(["localhost", "127.0.0.1", "::1", "172.*.*.*", "192.168.*.*", "10.0.*.*"])

ALLOWED_HOSTS = INTERNAL_IPS

DATA_UPLOAD_MAX_MEMORY_SIZE = 5000000

# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.sites',

    'debug_toolbar', # https://github.com/jazzband/django-debug-toolbar/

    # FIXME: https://bitbucket.org/kbr/autotask/pull-requests/3/
    # 'autotask', # https://bitbucket.org/kbr/autotask

    # 'easy_thumbnails', # https://github.com/SmileyChris/easy-thumbnails
    # 'treebeard', # https://github.com/django-treebeard/django-treebeard
    # 'sekizai', # https://github.com/ojii/django-sekizai
    # 'djangocms_text_ckeditor', # https://github.com/divio/djangocms-text-ckeditor
    # 'filer', # https://github.com/divio/django-filer

    'dynamic_fixtures', # https://github.com/Peter-Slump/django-dynamic-fixtures

    'for_runners',

    "for_runners_test_project.test_app",
)

from for_runners.app_settings import *  # @UnusedWildImport

# https://bitbucket.org/kbr/autotask
AUTOTASK_IS_ACTIVE = True

# Download map via geotiler in for_runners.gpx_tools.gpxpy2map.generate_map
MAP_DOWNLOAD = True
# MAP_DOWNLOAD = False

ROOT_URLCONF = 'for_runners_test_project.urls'
WSGI_APPLICATION = 'for_runners_test_project.wsgi.application'

MIDDLEWARE = (
    # https://github.com/jazzband/django-debug-toolbar/
    'debug_toolbar.middleware.DebugToolbarMiddleware',

    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.security.SecurityMiddleware',

    'django_tools.middlewares.ThreadLocal.ThreadLocalMiddleware',
)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, "templates/"),
        ],
        'OPTIONS': {
            'loaders': [
                ('django.template.loaders.cached.Loader', (
                    'django.template.loaders.filesystem.Loader',
                    'django.template.loaders.app_directories.Loader',
                )),
            ],
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.i18n',
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.template.context_processors.media',
                'django.template.context_processors.csrf',
                'django.template.context_processors.tz',
                'django.template.context_processors.static',

                "for_runners.context_processors.for_runners_version_string",
            ],
        },
    },
]


if DEBUG:
    # Disable caches:
    CACHES = {
        'default': {
            'BACKEND': 'django.core.cache.backends.dummy.DummyCache',
        }
    }
    # Disable CacheLoader:
    TEMPLATES[0]["OPTIONS"]["loaders"] = (
        'django.template.loaders.filesystem.Loader',
        'django.template.loaders.app_directories.Loader',
    )

# Database
# https://docs.djangoproject.com/en/1.8/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, "..", "test_project_db.sqlite3"),
        # 'NAME': ":memory:"
        # https://docs.djangoproject.com/en/dev/ref/databases/#database-is-locked-errors
        'timeout': 30,
    }
}


# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

# Default and fallback language:
# https://docs.djangoproject.com/en/1.11/ref/settings/#language-code
LANGUAGE_CODE = "en"

# http://django-parler.readthedocs.org/en/latest/quickstart.html#configuration
PARLER_LANGUAGES = {
    1: [
        {
            "name": _("German"),
            "code": "de",
            "fallbacks": [LANGUAGE_CODE],
            "hide_untranslated": False,
        },
        {
            "name": _("English"),
            "code": "en",
            "fallbacks": ["de"],
            "hide_untranslated": False,
        },
    ],
    "default": { # all SITE_ID"s
        "fallbacks": [LANGUAGE_CODE],
        "redirect_on_fallback": False,
    },
}


# https://docs.djangoproject.com/en/1.8/ref/settings/#languages
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGES = tuple([(d["code"], d["name"]) for d in PARLER_LANGUAGES[1]])

LANGUAGE_DICT = dict(LANGUAGES) # useful to get translated name by language code

# http://django-parler.readthedocs.org/en/latest/quickstart.html#configuration
PARLER_DEFAULT_LANGUAGE_CODE = LANGUAGE_CODE


USE_I18N = True

USE_L10N = True


# https://docs.djangoproject.com/en/1.11/topics/i18n/timezones/
# TODO: Use UTC and handle time zone
TIME_ZONE = "Europe/Paris"
USE_TZ = False
# TIME_ZONE = 'UTC'
# USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.8/howto/static-files/

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'static')

MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')


# https://django-debug-toolbar.readthedocs.io/en/stable/configuration.html#debug-toolbar-config
from debug_toolbar.settings import CONFIG_DEFAULTS as DEBUG_TOOLBAR_CONFIG

# don't load jquery from ajax.googleapis.com, just use django's version:
DEBUG_TOOLBAR_CONFIG["JQUERY_URL"] = STATIC_URL + "admin/js/vendor/jquery/jquery.min.js"


# Basic Django CMS settings

CMS_TEMPLATES = (
    ('base.html', 'Basic Page'),
    # ('tests/plugin_test.html', "Add CMS-Plugin test template")
)
CMS_PERMISSION = True

# Basic Placeholder config

# from djangocms_text_ckeditor.cms_plugins import TextPlugin
CKEDITOR = "TextPlugin"

CMS_PLACEHOLDER_CONF = {
    None: {
        'name': _("Content"),
        # Don't define 'plugin' to enable all existing CMS plugins
        # 'plugins': [
        #     CKEDITOR,
        #     plugin_anchor_menu_constants.ANCHOR_PLUGIN_NAME,
        #     plugin_anchor_menu_constants.DROP_DOWN_ANCHOR_MENU_PLUGIN_NAME,
        # ],
        'default_plugins': [
            {
                'plugin_type': CKEDITOR,
                'values': {'body': "Lorem ipsum dolor sit amet"},
            },
        ],
    },
}


# https://django-debug-toolbar.readthedocs.io/en/stable/configuration.html#debug-toolbar-config
from debug_toolbar.settings import CONFIG_DEFAULTS as DEBUG_TOOLBAR_CONFIG

# Disable some more panels that will slow down the page:
DEBUG_TOOLBAR_CONFIG["DISABLE_PANELS"].add('debug_toolbar.panels.sql.SQLPanel')
DEBUG_TOOLBAR_CONFIG["DISABLE_PANELS"].add('debug_toolbar.panels.cache.CachePanel')

# don't load jquery from ajax.googleapis.com, just use django's version:
DEBUG_TOOLBAR_CONFIG["JQUERY_URL"] = "/static/admin/js/vendor/jquery/jquery.min.js"

DEBUG_TOOLBAR_CONFIG["SHOW_COLLAPSED"] = True # Show toolbar collapsed by default.



#_____________________________________________________________________________
# cut 'pathname' in log output

import logging

old_factory = logging.getLogRecordFactory()


def cut_path(pathname, max_length):
    if len(pathname)<=max_length:
        return pathname
    return "...%s" % pathname[-(max_length-3):]


def record_factory(*args, **kwargs):
    record = old_factory(*args, **kwargs)
    record.cut_path = cut_path(record.pathname, 30)
    return record


logging.setLogRecordFactory(record_factory)


# -----------------------------------------------------------------------------


# https://docs.python.org/3/library/logging.html#logging-levels
LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'verbose': {
            'format': '%(levelname)8s %(cut_path)s:%(lineno)-3s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
        }
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': False,
        },
        "matplotlib": {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': False,
        },
        'django': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': False,
        },
        'django_tools': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': False,
        },
        'for_runners': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': False,
        },
    },
}



"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import io
import logging
import statistics
from pathlib import Path

from django.conf import settings
from django.core.files import File
from django.core.files.base import ContentFile
from django.db import models
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from django.utils.text import slugify
from django.utils.translation import ugettext_lazy as _

# https://github.com/jedie/django-tools
from django_tools.decorators import display_admin_error
from django_tools.models import UpdateTimeBaseModel

# https://github.com/jedie/django-for-runners
from for_runners.geo import reverse_geo
from for_runners.gpx import (
    GpxIdentifier, add_extension_data, get_2d_coordinate_list, get_extension_data, get_identifier, iter_distance,
    iter_points, parse_gpx
)
from for_runners.gpx_tools.humanize import human_distance, human_duration, human_seconds
from for_runners.managers.gpx import GpxModelManager
from for_runners.model_utils import ModelAdminUrlMixin
from for_runners.models import DistanceModel, ParticipationModel
from for_runners.svg import gpx2svg_file, gpx2svg_string
from for_runners.weather import NoWeatherData, meta_weather_com

log = logging.getLogger(__name__)


def svg_upload_path(instance, filename):
    # file will be uploaded to MEDIA_ROOT/user_<id>/<filename>
    return instance.get_svg_upload_path(filename)


class GpxModel(ModelAdminUrlMixin, UpdateTimeBaseModel):
    """
    inherit from ModelAdminUrlMixin:
        * get_admin_change_url()

    inherit from UpdateTimeBaseModel:
        * createtime
        * lastupdatetime
    """
    participation = models.ForeignKey(
        ParticipationModel,
        related_name="tracks",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )

    gpx = models.TextField(help_text="The raw gpx file content",)
    creator = models.CharField(
        help_text="Used device to create this track",
        max_length=511,
        null=True,
        blank=True,
    )
    track_svg = models.FileField(
        verbose_name=_("Track SVG"),
        upload_to=svg_upload_path,
        null=True,
        blank=True,
    )

    start_time = models.DateTimeField(
        editable=False,
        help_text=_("Start time of the first segment in track"),
    )
    start_latitude = models.FloatField(
        editable=False,
        help_text=_(
            "Latitude of the first recorded point from the *.gpx file"),
    )
    start_longitude = models.FloatField(
        editable=False,
        help_text=_(
            "Longitude of the first recorded point from the *.gpx file"),
    )
    start_temperature = models.FloatField(
        editable=True,
        null=True,
        blank=True,
        help_text=_("Temperature at start."),
    )
    start_weather_state = models.CharField(
        max_length=127,
        null=True,
        blank=True,
        help_text="Weather state at start.",
    )
    short_start_address = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="The short address of the start point",
    )
    full_start_address = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="The full address of the start point",
    )

    finish_time = models.DateTimeField(
        editable=False,
        help_text=_("End time of the last segment in track"),
    )
    finish_latitude = models.FloatField(
        editable=False,
        help_text=_("Latitude of the finish point"),
    )
    finish_longitude = models.FloatField(
        editable=False,
        help_text=_("Longitude of the finish point"),
    )
    finish_temperature = models.FloatField(
        editable=True,
        null=True,
        blank=True,
        help_text=_("Temperature at finish."),
    )
    finish_weather_state = models.CharField(
        max_length=127,
        null=True,
        blank=True,
        help_text="Weather state at finish.",
    )
    short_finish_address = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="The short address of the finish point",
    )
    full_finish_address = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="The full address of the finish point",
    )

    tracked_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        related_name="%(class)s_createby",
        null=True,
        blank=True,
        help_text="The user that tracked this gpx entry",
        on_delete=models.SET_NULL
    )
    lastupdateby = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        related_name="%(class)s_lastupdateby",
        null=True,
        blank=True,
        help_text="User as last edit this entry",
        on_delete=models.SET_NULL
    )

    points_no = models.PositiveIntegerField(
        help_text=_("Number of points in GPX"),
        null=True,
        blank=True,
    )

    length = models.PositiveIntegerField(
        help_text=
        _("Length in meters (calculated from GPX track 3-dimensional used latitude, longitude, and elevation)"
          ),
        null=True,
        blank=True,
    )
    ideal_distance = models.ForeignKey(
        to=DistanceModel,
        on_delete=models.SET_NULL,
        related_name="tracks",
        help_text=
        _("Length in meters (calculated 3-dimensional used latitude, longitude, and elevation)"
          ),
        null=True,
        blank=True,
    )

    duration_s = models.PositiveIntegerField(
        help_text=_("Duration in seconds (from the GPX track)"),
        editable=False,
        null=True,
        blank=True,
    )
    pace = models.DecimalField(
        help_text=_(
            "Min/km (number of minutes it takes to cover a kilometer)"),
        max_digits=4,
        decimal_places=2,  # store numbers up to 99 with a resolution of 2 decimal places
        null=True,
        blank=True,
    )

    uphill = models.IntegerField(
        help_text=_("Uphill elevation climbs in meters"),
        null=True,
        blank=True,
    )
    downhill = models.IntegerField(
        help_text=_("Downhill elevation descent in meters"),
        null=True,
        blank=True,
    )

    min_elevation = models.IntegerField(
        help_text=_("Minimum elevation in meters"),
        null=True,
        blank=True,
    )
    max_elevation = models.IntegerField(
        help_text=_("Maximum elevation in meters"),
        null=True,
        blank=True,
    )

    heart_rate_min = models.PositiveIntegerField(
        help_text=_("Minimum heart rate."),
        null=True,
        blank=True,
        editable=False)
    heart_rate_avg = models.PositiveIntegerField(
        help_text=_("Average heart rate."),
        null=True,
        blank=True,
        editable=False)
    heart_rate_max = models.PositiveIntegerField(
        help_text=_("Maximum heart rate."),
        null=True,
        blank=True,
        editable=False)

    objects = GpxModelManager()

    def save(self, *args, **kwargs):
        if self.gpx:
            self.calculate_values()

        super().save(*args, **kwargs)

        # TODO: schedule request weather info, if not set

    def get_svg_upload_path(self, filename):
        date_prefix = self.start_time.strftime("%Y_%m")
        svg_upload_path = "track_svg_%s/%s.svg" % (date_prefix, self.get_prefix_id())
        log.debug("Ignore source filename: %r upload to: %r", filename, svg_upload_path)
        return svg_upload_path

    def svg_tag(self):
        if self.track_svg:
            html = '<img src="{}" alt="gpx track" height="70px" width="70px" />'.format(self.track_svg.url)
            return mark_safe(html)
        return ""

    svg_tag.short_description = _("SVG")

    def svg_tag_big(self):
        if self.track_svg:
            html = '<img src="{}" alt="gpx track" height="200px" width="200px" />'.format(self.track_svg.url)
            return mark_safe(html)
        return ""

    svg_tag_big.short_description = _("SVG")

    def start_end_address(self):
        if self.short_start_address == self.short_finish_address:
            return "\u27F3 %s" % self.short_start_address
        html = "%s<br>\u25BE<br>%s" % (self.short_start_address, self.short_finish_address)
        return mark_safe(html)

    start_end_address.short_description = _("Start/End Address")

    def get_ideal_ratio(self):
        if self.ideal_distance:
            ratio = (float(self.ideal_distance.distance_km) * 1000) / self.length
            return ratio

    def get_ideal_duration_s(self):
        ratio = self.get_ideal_ratio()
        if ratio:
            return self.duration_s * ratio

    def get_ideal_pace(self):
        ratio = self.get_ideal_ratio()
        if ratio:
            return self.pace * ratio

    def get_ideal_distance_diff_m(self):
        if self.ideal_distance:
            distance_diff_m = (float(self.ideal_distance.distance_km) * 1000) - self.length
            return distance_diff_m

    def human_ideal_length(self):
        """
        used as labels in chart.js
        """
        if self.ideal_distance:
            return self.ideal_distance.get_human_distance()
        return human_distance(self.length / 1000)

    def human_length(self):
        if self.length:
            if self.ideal_distance:
                return self.ideal_distance.get_human_distance()
            else:
                return human_distance(self.length / 1000)

    def human_length_html(self):
        """
        Enhanced version of self.human_length()
        with more information via <span title="...">
        """
        if self.length:
            length_km = self.length / 1000

            if self.ideal_distance:
                diff_km = abs(self.get_ideal_distance_diff_m() / 1000)
                html = (
                    '<span'
                    ' title="{ideal} is the standardized distance'
                    ' - GPX-Track: {gpx} (diff: {diff})"'
                    '>{ideal}</span>'
                ).format(
                    ideal=self.ideal_distance.get_human_distance(),
                    gpx=human_distance(length_km),
                    diff=human_distance(diff_km)
                )
            else:
                html = ('<span title="real distance">%s</span>') % human_distance(length_km)

            return mark_safe(html)

    human_length_html.short_description = _("Length")
    human_length_html.admin_order_field = "length"

    def get_net_duration_s(self):
        if self.participation:
            # Use net duration from event participation
            return self.participation.get_duration_s()

    def human_duration(self):
        # Use net duration from event participation
        net_duration_s = self.get_net_duration_s()
        if net_duration_s is not None:
            return human_seconds(net_duration_s)

        ideal_duration_s = self.get_ideal_duration_s()
        if ideal_duration_s:
            return human_seconds(ideal_duration_s)

        if self.duration_s:
            return human_seconds(self.duration_s)

    def human_duration_html(self):
        """
        Enhanced version of self.human_duration()
        with more information via <span title="...">
        """
        net_duration_s = self.get_net_duration_s()
        if net_duration_s is not None:
            duration_diff = self.duration_s - net_duration_s
            html = (
                '<span'
                ' title="{net} is the official net duration'
                ' - GPX-Track: {gpx} (diff: {diff})"'
                '>{net}</span>'
            ).format(
                net=human_seconds(net_duration_s),
                gpx=human_seconds(self.duration_s),
                diff=human_duration(duration_diff),
            )
            return mark_safe(html)

        ideal_duration_s = self.get_ideal_duration_s()
        if ideal_duration_s:
            duration_diff = self.duration_s - ideal_duration_s
            html = (
                '<span'
                ' title="{ideal} is the standardized duration'
                ' - GPX-Track: {gpx} (diff: {diff})"'
                '>{ideal}</span>'
            ).format(
                ideal=human_seconds(ideal_duration_s),
                gpx=human_seconds(self.duration_s),
                diff=human_duration(duration_diff),
            )
            return mark_safe(html)

        if self.duration_s:
            html = ('<span' ' title="real duration"' '>%s</span>') % human_seconds(self.duration_s)
            return mark_safe(html)

    human_duration_html.short_description = _("Duration")
    human_duration_html.admin_order_field = "duration_s"

    def human_pace(self):
        if self.pace:
            return "%s min/km" % human_seconds(self.pace * 60)

    human_pace.short_description = _("Pace")
    human_pace.admin_order_field = "pace"

    def human_weather(self):
        if not self.start_temperature:
            return "-"
        html = "%s°C<br/>%s" % (round(self.start_temperature, 1), self.start_weather_state)
        return mark_safe(html)

    human_weather.short_description = _("Weather")
    human_weather.admin_order_field = "start_temperature"

    def _coordinate2link(self, lat, lon):
        html = (
            '<a'
            ' href="https://nominatim.openstreetmap.org/reverse?lat={lat}&lon={lon}&format=json&addressdetails=1"'
            ' title="Reverse {lat},{lon} address with OpenStreepMap"'
            ' target="_blank"'
            '>reverse address</a>'
            '<br>'
            '<a'
            ' href="https://www.openstreetmap.org/search?query={lat}%2C{lon}"'
            ' title="OpenStreepMap at {lat},{lon}"'
            ' target="_blank"'
            '>map</a>'
        ).format(
            lat=lat, lon=lon
        )
        return mark_safe(html)

    def start_coordinate_html(self):
        """
        return HTML Links for start point.
        """
        if self.start_latitude and self.start_longitude:
            return self._coordinate2link(
                lat=self.start_latitude,
                lon=self.start_longitude,
            )

    start_coordinate_html.short_description = _("Start coordinates")

    def finish_coordinate_html(self):
        """
        return HTML Links for finish point.
        """
        if self.finish_latitude and self.finish_longitude:
            return self._coordinate2link(
                lat=self.finish_latitude,
                lon=self.finish_longitude,
            )

    finish_coordinate_html.short_description = _("finish coordinates")

    def point_density(self):
        """
        Calculate the "density" of the GPX signal:
        :return: float - Meters/point count
        """
        if self.length > 0:
            density = self.length / self.points_no
            return density

    def gpx_meta(self):
        gpxpy_instance = self.get_gpxpy_instance()
        attr_names = (
            "version",
            "creator",
            "name",
            "description",
            "author_name",
            "author_email",
            "author_link",
            "author_link_text",
            "author_link_type",
            "copyright_author",
            "copyright_year",
            "copyright_license",
            "link",
            "link_text",
            "link_type",
            "time",
            "keywords",
        )
        result = []
        for attr_name in attr_names:
            value = getattr(gpxpy_instance, attr_name, None)
            if value:
                result.append((attr_name, value))

        return result

    _GPXPY_CACHE = {}

    def get_gpxpy_instance(self):
        try:
            return self._GPXPY_CACHE[self.pk]
        except KeyError:
            if self.gpx:
                gpxpy_instance = parse_gpx(content=self.gpx)
                if self.pk is not None:
                    self._GPXPY_CACHE[self.pk] = gpxpy_instance
                return gpxpy_instance

    def calc_pace(self):
        duration_s = self.get_net_duration_s()
        if not duration_s:
            duration_s = self.get_ideal_duration_s()
        if not duration_s:
            duration_s = self.duration_s
        if not duration_s:
            return None

        if self.ideal_distance:
            distance_km = int(self.ideal_distance.distance_km)
        else:
            distance_km = self.length / 1000

        try:
            pace = (duration_s / 60) / distance_km
        except ZeroDivisionError:
            # FIXME
            log.exception("Error calculate pace with duration %rsec and distance %skm!" % (duration_s, distance_km))
        else:
            if pace > 99 or pace < 0:
                log.error("Pace out of range: %f", pace)
            else:
                self.pace = pace

    def calculate_values(self):
        if not self.gpx:
            return

        gpxpy_instance = self.get_gpxpy_instance()
        self.points_no = gpxpy_instance.get_points_no()
        self.length = gpxpy_instance.length_3d()

        try:
            ideal_distances_qs = DistanceModel.objects.filter(
                min_distance_m__lte=self.length,
                max_distance_m__gte=self.length,
            )
        except DistanceModel.DoesNotExist:
            pass
        else:
            ideal_distance_count = ideal_distances_qs.count()
            if ideal_distance_count > 1:
                log.error("Found more the one ideal distances for %i Meters", self.length)
                ideal_distances_qs = ideal_distances_qs.order_by("distance_km")
                ideal_distance = ideal_distances_qs[0]
            elif ideal_distance_count == 1:
                ideal_distance = ideal_distances_qs.get()
            else:
                ideal_distance = None

            if ideal_distance:
                self.ideal_distance = ideal_distance
                log.debug("Set ideal distance to %s", self.ideal_distance)

        # e.g: GPX without a track return 0
        duration = gpxpy_instance.get_duration()
        if duration:
            self.duration_s = duration
            self.calc_pace()

        uphill_downhill = gpxpy_instance.get_uphill_downhill()
        self.uphill = uphill_downhill.uphill
        self.downhill = uphill_downhill.downhill

        elevation_extremes = gpxpy_instance.get_elevation_extremes()
        self.min_elevation = elevation_extremes.minimum
        self.max_elevation = elevation_extremes.maximum

        identifier = get_identifier(gpxpy_instance)

        self.start_time = identifier.start_time
        self.finish_time = identifier.finish_time
        self.start_latitude = identifier.start_lat
        self.start_longitude = identifier.start_lon
        self.finish_latitude = identifier.finish_lat
        self.finish_longitude = identifier.finish_lon

        if not self.start_temperature:
            try:
                temperature, weather_state = meta_weather_com.coordinates2weather(
                    self.start_latitude, self.start_longitude, date=self.start_time, max_seconds=self.duration_s
                )
            except NoWeatherData:
                log.error("No weather data for start.")
            else:
                self.start_temperature = temperature
                self.start_weather_state = weather_state

        if not self.finish_temperature:
            try:
                temperature, weather_state = meta_weather_com.coordinates2weather(
                    self.finish_latitude, self.finish_longitude, date=self.finish_time, max_seconds=self.duration_s
                )
            except NoWeatherData:
                log.error("No weather data for finish.")
            else:
                self.finish_temperature = temperature
                self.finish_weather_state = weather_state

        if not self.full_start_address:
            try:
                start_address = reverse_geo(self.start_latitude, self.start_longitude)
            except Exception as err:
                # e.g.: geopy.exc.GeocoderTimedOut: Service timed out
                log.error("Can't reverse geo: %s" % err)
            else:
                self.short_start_address = start_address.short
                self.full_start_address = start_address.full

        if not self.full_finish_address:
            try:
                finish_address = reverse_geo(self.finish_latitude, self.finish_longitude)
            except Exception as err:
                # e.g.: geopy.exc.GeocoderTimedOut: Service timed out
                log.error("Can't reverse geo: %s" % err)
            else:
                self.short_finish_address = finish_address.short
                self.full_finish_address = finish_address.full

        if not self.track_svg:
            log.debug("Create SVG from GPX...")
            svg_string = gpx2svg_string(gpxpy_instance)
            content = ContentFile(svg_string)

            # https://docs.djangoproject.com/en/2.0/ref/models/fields/#django.db.models.fields.files.FieldFile.save
            self.track_svg.save(
                name="temp.svg",  # real file path will be set in self.get_svg_upload_path()
                content=content,
                save=False
            )

        # TODO: Handle other extensions, too.
        # Garmin containes also 'cad'
        extension_data = get_extension_data(gpxpy_instance)
        if extension_data is not None and "hr" in extension_data:
            heart_rates = extension_data["hr"]
            self.heart_rate_min = min(heart_rates)
            self.heart_rate_avg = statistics.median(heart_rates)
            self.heart_rate_max = max(heart_rates)

        if not self.creator:
            self.creator = gpxpy_instance.creator

    def short_name(self):
        if self.pk is None:
            return "new, unsaved GPX Track"

        parts = [self.start_time.strftime("%Y-%m-%d")]
        if self.participation:
            parts.append(self.participation.event.name)
        else:
            parts.append(self.short_start_address)
        result = " ".join([str(part) for part in parts if part])
        if result:
            return result
        return "GPX Track ID:%s" % self.pk

    def get_short_slug(self):
        name = self.short_name()
        return slugify(name)

    def get_prefix_id(self):
        gpxpy_instance = self.get_gpxpy_instance()
        prefix_id = GpxIdentifier(gpxpy_instance).get_prefix_id()
        return prefix_id

    def __str__(self):
        # return self.get_prefix_id()
        return self.short_name()

    class Meta:
        verbose_name = _('GPX Track')
        verbose_name_plural = _('GPX Tracks')
        unique_together = ((
            "start_time", "start_latitude", "start_longitude", "finish_time", "finish_latitude", "finish_longitude"
        ),)
        ordering = ('-start_time', '-pk')



# Generated by Django 2.0.7 on 2018-07-31 18:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('for_runners', '0004_auto_20180719_1739'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='gpxmodel',
            name='net_duration',
        ),
        migrations.AlterField(
            model_name='gpxmodel',
            name='duration',
            field=models.PositiveIntegerField(blank=True, help_text='Duration in seconds (from the GPX track)', null=True),
        ),
        migrations.AlterField(
            model_name='gpxmodel',
            name='length',
            field=models.PositiveIntegerField(blank=True, help_text='Length in meters (calculated from GPX track 3-dimensional used latitude, longitude, and elevation)', null=True),
        ),
        migrations.RenameField(
            model_name='gpxmodel',
            old_name='duration',
            new_name='duration_s',
        ),
        migrations.AlterField(
            model_name='gpxmodel',
            name='duration_s',
            field=models.PositiveIntegerField(blank=True, editable=False, help_text='Duration in seconds (from the GPX track)', null=True),
        ),
    ]



# coding: utf-8

import os
from unittest import TestCase

import for_runners_test_project
from django_tools.unittest_utils.django_command import DjangoCommandMixin

MANAGE_DIR = os.path.abspath(os.path.dirname(for_runners_test_project.__file__))


class CheckTestEnvironment(DjangoCommandMixin, TestCase):

    def call_manage_py(self, *args, **kwargs):
        return super(CheckTestEnvironment, self).call_manage_py(args, manage_dir=MANAGE_DIR, **kwargs)

    def test_help(self):
        output = self.call_manage_py("--help")
        print(output)
        self.assertNotIn("ERROR", output)
        self.assertIn("[django]", output)
        self.assertIn("Type 'manage.py help <subcommand>' for help on a specific subcommand.", output)

    def test_django_check(self):
        output = self.call_manage_py("check")
        self.assertIn("System check identified no issues (0 silenced).", output)



# Generated by Django 2.0.7 on 2018-07-19 17:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('for_runners', '0003_auto_20180712_2043'),
    ]

    operations = [
        migrations.AlterField(
            model_name='costmodel',
            name='name',
            field=models.CharField(help_text='The name of the item you dis pay for', max_length=15),
        ),
    ]



"""
    Django-ForRunners App settings
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    include in your own settings.py e.g.:

        from for_runners.app_settings import *

    created 28.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import decimal

# Base data for:
#   for_runners.models.distance.DistanceModel
#
# Can be change via:
#   settings.BASE_IDEAL_TRACK_LENGTHS
#
# Used in
#   for_runners.management.commands.fill_basedata.Command
#
BASE_IDEAL_TRACK_LENGTHS=( # All values are in kilometers ;)
    decimal.Decimal(0.4),
    decimal.Decimal(0.8),
    decimal.Decimal(5.0),
    decimal.Decimal(6.5),
    decimal.Decimal(7.5),
    decimal.Decimal(8.5),
    decimal.Decimal(10.0),
    decimal.Decimal(21.0975),
    decimal.Decimal(42.195),
)

# All cash values are stored as decimal field without any currency information
# This symbol will be just added ;)
FOR_RUNNERS_CURRENCY_SYMBOL="€"



#!/usr/bin/env python3
# coding: utf-8

from __future__ import unicode_literals, print_function, absolute_import

import os
import shutil
import sys
import distutils
import subprocess

from setuptools import setup, find_packages


def read(*args):
    return open(os.path.join(os.path.dirname(__file__), *args)).read()


class BaseCommand(distutils.cmd.Command):
    user_options = []
    def initialize_options(self): pass
    def finalize_options(self): pass


class ToxTestCommand(BaseCommand):
    """Distutils command to run tests via tox: 'python setup.py tox'."""
    description = "Run tests via 'tox'."

    def run(self):
        self.announce("Running tests with 'tox'...", level=distutils.log.INFO)
        returncode = subprocess.call(['tox'])
        sys.exit(returncode)


class TestCommand(BaseCommand):
    """Distutils command to run tests via py.test: 'python setup.py test'."""
    description = "Run tests via 'py.test'."

    def run(self):
        self.announce("Running tests...", level=distutils.log.INFO)
        returncode = subprocess.call(['pytest'])
        sys.exit(returncode)


__version__="<unknown>"
exec(read('for_runners', 'version.py'))


PACKAGE_ROOT = os.path.os.path.dirname(os.path.abspath(__file__))


#_____________________________________________________________________________
# convert creole to ReSt on-the-fly, see also:
# https://github.com/jedie/python-creole/wiki/Use-In-Setup
long_description = None
for arg in ("test", "check", "register", "sdist", "--long-description"):
    if arg in sys.argv:
        try:
            from creole.setup_utils import get_long_description
        except ImportError as err:
            raise ImportError("%s - Please install python-creole - e.g.: pip install python-creole" % err)
        else:
            long_description = get_long_description(PACKAGE_ROOT)
        break
#----------------------------------------------------------------------------


if "publish" in sys.argv:
    """
    'publish' helper for setup.py

    Build and upload to PyPi, if...
        ... __version__ doesn't contains "dev"
        ... we are on git 'master' branch
        ... git repository is 'clean' (no changed files)

    Upload with "twine", git tag the current version and git push --tag

    The cli arguments will be pass to 'twine'. So this is possible:
     * Display 'twine' help page...: ./setup.py publish --help
     * use testpypi................: ./setup.py publish --repository=test

    TODO: Look at: https://github.com/zestsoftware/zest.releaser

    Source: https://github.com/jedie/python-code-snippets/blob/master/CodeSnippets/setup_publish.py
    copyleft 2015-2017 Jens Diemer - GNU GPL v2+
    """
    if sys.version_info[0] == 2:
        input = raw_input

    import_error = False
    try:
        # Test if wheel is installed, otherwise the user will only see:
        #   error: invalid command 'bdist_wheel'
        import wheel
    except ImportError as err:
        print("\nError: %s" % err)
        print("\nMaybe https://pypi.python.org/pypi/wheel is not installed or virtualenv not activated?!?")
        print("e.g.:")
        print("    ~/your/env/$ source bin/activate")
        print("    ~/your/env/$ pip install wheel")
        import_error = True

    try:
        import twine
    except ImportError as err:
        print("\nError: %s" % err)
        print("\nMaybe https://pypi.python.org/pypi/twine is not installed or virtualenv not activated?!?")
        print("e.g.:")
        print("    ~/your/env/$ source bin/activate")
        print("    ~/your/env/$ pip install twine")
        import_error = True

    if import_error:
        sys.exit(-1)

    def verbose_check_output(*args):
        """ 'verbose' version of subprocess.check_output() """
        call_info = "Call: %r" % " ".join(args)
        try:
            output = subprocess.check_output(args, universal_newlines=True, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as err:
            print("\n***ERROR:")
            print(err.output)
            raise
        return call_info, output

    def verbose_check_call(*args):
        """ 'verbose' version of subprocess.check_call() """
        print("\tCall: %r\n" % " ".join(args))
        subprocess.check_call(args, universal_newlines=True)

    def confirm(txt):
        print("\n%s" % txt)
        if input("\nPublish anyhow? (Y/N)").lower() not in ("y", "j"):
            print("Bye.")
            sys.exit(-1)

    if "dev" in __version__:
        confirm("WARNING: Version contains 'dev': v%s\n" % __version__)

    print("\nCheck if we are on 'master' branch:")
    call_info, output = verbose_check_output("git", "branch", "--no-color")
    print("\t%s" % call_info)
    if "* master" in output:
        print("OK")
    else:
        confirm("\nNOTE: It seems you are not on 'master':\n%s" % output)

    print("\ncheck if if git repro is clean:")
    call_info, output = verbose_check_output("git", "status", "--porcelain")
    print("\t%s" % call_info)
    if output == "":
        print("OK")
    else:
        print("\n *** ERROR: git repro not clean:")
        print(output)
        sys.exit(-1)

    print("\ncheck if pull is needed")
    verbose_check_call("git", "fetch", "--all")
    call_info, output = verbose_check_output("git", "log", "HEAD..origin/master", "--oneline")
    print("\t%s" % call_info)
    if output == "":
        print("OK")
    else:
        print("\n *** ERROR: git repro is not up-to-date:")
        print(output)
        sys.exit(-1)
    verbose_check_call("git", "push")

    print("\nRun './setup.py check':")
    call_info, output = verbose_check_output("./setup.py", "check")
    if "warning" in output:
        print(output)
        confirm("Warning found!")
    else:
        print("OK")

    print("\nCleanup old builds:")
    def rmtree(path):
        path = os.path.abspath(path)
        if os.path.isdir(path):
            print("\tremove tree:", path)
            shutil.rmtree(path)
    rmtree("./dist")
    rmtree("./build")

    print("\nbuild but don't upload...")
    log_filename="build.log"
    with open(log_filename, "a") as log:
        call_info, output = verbose_check_output(
            sys.executable or "python",
            "setup.py", "sdist", "bdist_wheel", "bdist_egg"
        )
        print("\t%s" % call_info)
        log.write(call_info)
        log.write(output)
    print("Build output is in log file: %r" % log_filename)

    git_tag="v%s" % __version__

    print("\ncheck git tag")
    call_info, output = verbose_check_output("git", "log", "HEAD..origin/master", "--oneline")
    if git_tag in output:
        print("\n *** ERROR: git tag %r already exists!" % git_tag)
        print(output)
        sys.exit(-1)
    else:
        print("OK")

    print("\nUpload with twine:")
    twine_args = sys.argv[1:]
    twine_args.remove("publish")
    twine_args.insert(1, "dist/*")
    print("\ttwine upload command args: %r" % " ".join(twine_args))
    from twine.commands.upload import main as twine_upload
    twine_upload(twine_args)

    print("\ngit tag version")
    verbose_check_call("git", "tag", git_tag)

    print("\ngit push tag to server")
    verbose_check_call("git", "push", "--tags")

    sys.exit(0)


# https://pypi.org/classifiers/
classifiers = """
Development Status :: 3 - Alpha
#Development Status :: 4 - Beta
#Development Status :: 5 - Production/Stable
#Development Status :: 6 - Mature
#Development Status :: 7 - Inactive  
Intended Audience :: Developers
Intended Audience :: End Users/Desktop
License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)
Operating System :: POSIX
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3 :: Only
Framework :: Django
Framework :: Django :: 2.1
Topic :: Internet
"""


setup(
    name='django-for-runners',
    version=__version__,
    description='Store your GPX tracks of your running (or other sports activity) in django.',
    long_description=long_description,
    author='Jens Diemer',
    author_email='django-for-runners@jensdiemer.de',
    url='https://github.com/jedie/django-for-runners',
    license="GNU General Public License v3.0 or above",
    classifiers=[c.strip() for c in classifiers.splitlines()
                 if c.strip() and not c.startswith('#')],
    packages=find_packages(),
    python_requires='>=3.5',

    install_requires=["django>2.1",],

    include_package_data=True,
    cmdclass={
        'test': TestCommand,
        'tox': ToxTestCommand,
    }
)



"""
    created 30.05.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging

from django.db import models
from django.utils.translation import ugettext_lazy as _

# https://github.com/jedie/django-for-runners
from for_runners.gpx_tools.humanize import human_distance

log = logging.getLogger(__name__)


class DistanceModel(models.Model):
    distance_km = models.DecimalField(
        help_text=_("The ideal track length in kilometer."),
        unique=True,
        # store numbers up to 999 with a resolution of 4 decimal places
        max_digits=7,
        decimal_places=4,
    )
    variance = models.PositiveSmallIntegerField(
        help_text=_(
            "Maximum (+/-) deviation in percent to match this distance."),
        default=5)
    min_distance_m = models.PositiveIntegerField(
        editable=False,
        blank=True,
        null=True,
    )
    max_distance_m = models.PositiveIntegerField(
        editable=False,
        blank=True,
        null=True,
    )

    def _set_min_max(self):
        distance_m = self.distance_km * 1000
        variance_m = self.get_variance_m()
        self.min_distance_m = round(distance_m - variance_m)
        self.max_distance_m = round(distance_m + variance_m)

    def save(self, *args, **kwargs):
        self._set_min_max()
        log.debug("Save: %s", self)
        super().save(*args, **kwargs)

    def get_variance_m(self):
        variance_km = self.distance_km / 100 * self.variance
        return variance_km * 1000

    def get_human_distance(self):
        return human_distance(self.distance_km)

    get_human_distance.short_description = _("Distance")
    get_human_distance.admin_order_field = "distance_km"

    def get_human_variance(self):
        return "%i %%" % self.variance

    get_human_variance.short_description = _("Variance")

    def get_human_variance_as_length(self):
        return human_distance(self.get_variance_m() / 1000)

    get_human_variance_as_length.short_description = _("Variance")

    def get_human_min_max(self):
        return "%s - %s" % (human_distance(self.min_distance_m / 1000), human_distance(self.max_distance_m / 1000))

    get_human_min_max.short_description = _("Min/Max")

    def __str__(self):
        return self.get_human_distance()

    class Meta:
        verbose_name = _('Distance')
        verbose_name_plural = _('Distances')
        ordering = ('-distance_km',)



# coding: utf-8

import unittest

import pytest

# https://github.com/jedie/django-tools
from django_tools.unittest_utils.selenium_utils import (
    SeleniumChromiumTestCase, SeleniumFirefoxTestCase, chromium_available, firefox_available
)
from django_tools.unittest_utils.unittest_base import BaseTestCase
from django_tools.unittest_utils.user import TestUserMixin

# https://github.com/jedie/django-for-runners
from for_runners.version import __version__


@pytest.mark.django_db
class AdminAnonymousTests(BaseTestCase):
    """
    Anonymous will be redirected to the login page.
    """

    def test_login_en(self):
        response = self.client.get('/en/admin/', HTTP_ACCEPT_LANGUAGE='en')
        self.assertRedirects(response, expected_url='/en/admin/login/?next=/en/admin/')

    def test_login_de(self):
        response = self.client.get('/de/admin/', HTTP_ACCEPT_LANGUAGE='de')
        self.assertRedirects(response, expected_url='/de/admin/login/?next=/de/admin/')


@pytest.mark.django_db
class AdminLoggedinTests(TestUserMixin, AdminAnonymousTests):
    """
    Some basics test with the django admin
    """

    def test_staff_admin_index(self):
        self.login(usertype='staff')
        response = self.client.get('/en/admin/', HTTP_ACCEPT_LANGUAGE='en')
        self.assertResponse(
            response,
            must_contain=(
                "<title>Site administration | Django-ForRunners v%s</title>" % __version__,
                "<h1>Site administration</h1>",
                "<strong>staff_test_user</strong>",
                "<p>You don't have permission to view or edit anything.</p>",
            ),
            must_not_contain=('error', 'traceback', "/add/", "/change/"),
            template_name='admin/index.html',
            messages=[],
            html=True
        )

    def test_superuser_admin_index(self):
        self.login(usertype='superuser')
        response = self.client.get('/en/admin/', HTTP_ACCEPT_LANGUAGE='en')
        self.assertResponse(
            response,
            must_contain=(
                'Django-ForRunners',
                'superuser',
                'Site administration',
                '/admin/auth/group/add/',
                '/admin/auth/user/add/',
            ),
            must_not_contain=('error', 'traceback'),
            template_name='admin/index.html',
        )


@unittest.skipUnless(chromium_available(), "Skip because Chromium is not available!")
class AdminChromiumTests(SeleniumChromiumTestCase):

    def test_admin_login_page(self):
        self.driver.get(self.live_server_url + "/admin/login/")
        self.assert_equal_page_title("Log in | Django-ForRunners v%s" % __version__)
        self.assert_in_page_source('<form action="/en/admin/login/" method="post" id="login-form">')
        self.assert_no_javascript_alert()


@unittest.skipUnless(firefox_available(), "Skip because Firefox is not available!")
class AdminFirefoxTests(SeleniumFirefoxTestCase):

    def test_admin_login_page(self):
        self.driver.get(self.live_server_url + "/admin/login/")
        self.assert_equal_page_title("Log in | Django-ForRunners v%s" % __version__)
        self.assert_in_page_source('<form action="/en/admin/login/" method="post" id="login-form">')
        self.assert_no_javascript_alert()



"""
    API connection to metaweather.com

    https://www.metaweather.com/api/#locationday
    https://www.metaweather.com/api/location/search/?lattlong=36.96,-122.02
    https://www.metaweather.com/api/location/2487956/2013/4/30/

    created 21.06.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018 by the django-for-runners team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import collections
import datetime
import json
import logging
import socket
import statistics
import time
import urllib
import urllib.request
from pprint import pprint
from urllib.error import URLError

log = logging.getLogger(__name__)


class NoWeatherData(ValueError):
    pass


def request_json(url, timeout=3, user_agent="python"):
    request = urllib.request.Request(url)
    request.add_header('User-Agent', user_agent)

    start_time = time.time()
    try:
        with urllib.request.urlopen(request, timeout=timeout) as f:
            json_data = f.read()
    except (URLError, socket.timeout) as err:
        print("ERROR: %s (url: %r)" % (err, url))
        raise NoWeatherData
    else:
        response_time_ms = round((time.time() - start_time) * 1000, 1)
        print("Response in: %.1fms (url: %r)" % (response_time_ms, url))
        if isinstance(json_data, bytes):
            json_data = json_data.decode("utf-8")
        data = json.loads(json_data)
        return data


class MetaWeatherCom:

    API_URL = "https://www.metaweather.com/api"

    def _request(self, url):
        url = self.API_URL + url
        data = request_json(url)
        return data

    _lat_lon2woeid_cache = {}

    def lat_lon2woeid(self, lat, lon, decimal_places=2):
        """
        lat/lon coordinates to WOEID (Where On Earth ID)

        :return: nearest woeid, json_data
        """
        lat = round(lat, decimal_places)
        lon = round(lon, decimal_places)

        try:
            result = self._lat_lon2woeid_cache[(lat, lon)]
            log.info("Fetch WOEID from cache!")
            return result
        except KeyError:
            log.info("Request WOEID")

        json_data = self._request(url="/location/search/?lattlong=%s,%s" % (lat, lon))
        # pprint(json_data)
        # [{'distance': 1836,
        #   'latt_long': '36.974018,-122.030952',
        #   'location_type': 'City',
        #   'title': 'Santa Cruz',
        #   'woeid': 2488853},
        #  {'distance': 43722,
        #   'latt_long': '37.338581,-121.885567',
        #   'location_type': 'City',
        #   'title': 'San Jose',
        #   'woeid': 2488042},
        #  {'distance': 49177,
        #   'latt_long': '37.39999,-122.079552',
        #   'location_type': 'City',
        #   'title': 'Mountain View',
        #   'woeid': 2455920},
        #  ...]
        last_distance = None
        nearest_woeid = None
        woe_data = None
        for result in json_data:
            distance = int(result["distance"])
            if last_distance is None or distance < last_distance:
                last_distance = distance
                woe_data = result
                nearest_woeid = int(woe_data["woeid"])

        result = (nearest_woeid, woe_data, json_data)
        self._lat_lon2woeid_cache[(lat, lon)] = result
        return result

    _location_day_cache = {}

    def location_day(self, woeid, date, max_seconds=60):

        url = "/location/{woeid}/{year}/{month}/{day}/".format(
            woeid=woeid, year=date.year, month=date.month, day=date.day
        )
        try:
            json_data = self._location_day_cache[url]
        except KeyError:
            log.debug("Request %s...", url)
            json_data = self._request(url)
            self._location_day_cache[url] = json_data
        else:
            log.debug("Fetch %s from cache!", url)

        if not json_data:
            log.error("json response is empty")
            raise NoWeatherData("json response is empty")

        # pprint(json_data)
        # [{'air_pressure': 1005.36,
        # 'applicable_date': '2017-04-30',
        # 'created': '2017-05-01T07:19:02.690650Z',
        # 'humidity': 55,
        # 'id': 6301075443286016,
        # 'max_temp': 23.453333333333333,
        # 'min_temp': 10.014999999999999,
        # 'predictability': 70,
        # 'the_temp': 23.97666666666667,
        # 'visibility': 12.81857701026008,
        # 'weather_state_abbr': 'lc',
        # 'weather_state_name': 'Light Cloud',
        # 'wind_direction': 305.4904669344955,
        # 'wind_direction_compass': 'NW',
        # 'wind_speed': 5.540785988894192},
        # ...]
        data = collections.defaultdict(list)
        for result in json_data:
            created_string = result["created"]  # e.g.: 2018-06-20T22:30:29.227960Z
            created_datetime = datetime.datetime.strptime(created_string, "%Y-%m-%dT%H:%M:%S.%fZ")

            delta = abs(date - created_datetime)
            data[delta.total_seconds()] = result

        temperatures = []
        weather_state_counter = collections.Counter()
        for delta_seconds, result in sorted(data.items()):
            if len(temperatures) > 0 and delta_seconds > max_seconds:
                # Skip if "created" date time is more than max_seconds, but collect at least one result
                continue

            temperature = statistics.median([result["min_temp"], result["the_temp"], result["max_temp"]])
            log.debug("Collect %.1f°C from %s (delta: %isec)", temperature, result["created"], delta_seconds)
            temperatures.append(temperature)

            weather_state_counter[result["weather_state_name"]] += 1

            # print("%s°C" % round(temperature), created_datetime, delta, delta_hours)

        temperature = statistics.median(temperatures)

        # print(weather_state_counter) # e.g.: Counter({'Light Cloud': 5, 'Showers': 2, 'Heavy Cloud': 1})
        weather_states = [item[0] for item in weather_state_counter.most_common()
                         ]  # e.g.: ['Light Cloud', 'Showers', 'Heavy Cloud']
        # print(weather_states)
        weather_state = "/".join(weather_states[:2])  # e.g.: Light Cloud/Showers

        log.info("Result: %.1f°C %r", temperature, weather_state)

        return temperature, weather_state

        #
        # print(
        #     created_datetime,
        #     "%s°C" % round(result["min_temp"]),
        #     "%s°C" % round(result["the_temp"]),
        #     "%s°C" % round(result["max_temp"]),
        #     "-> %s°C" % round(temp),
        #     result["predictability"],
        #     result["weather_state_name"],
        # )

    def coordinates2weather(self, lat, lon, date, max_seconds=12):
        if date is None:
            raise NoWeatherData("Can't get weather if date is None!")

        nearest_woeid, woe_data, json_data = self.lat_lon2woeid(lat, lon)
        log.info("Use nearest WOEID: %i (%r)", nearest_woeid, woe_data)
        temperature, weather_state = self.location_day(woeid=nearest_woeid, date=date, max_seconds=max_seconds)
        return temperature, weather_state


meta_weather_com = MetaWeatherCom()

if __name__ == "__main__":
    # Duisburg:
    # https://www.metaweather.com/api/location/search/?lattlong=51.4109,6.7828
    # woeid=648820 # Essen, city

    # nearest_woeid, woe_data, json_data = MetaWeatherCom().lat_lon2woeid(36.96, -122.02)
    # print("woeid:", nearest_woeid, woe_data) # woeid: 2488853 {'distance': 1836, 'title': 'Santa Cruz', 'location_type': 'City', 'woeid': 2488853, 'latt_long': '36.974018,-122.030952'}

    # /api/location/2487956/2013/4/30/ - San Francisco on 30th April 2013
    # MetaWeatherCom().location_day(woeid=2488853, date=datetime.date(year=2017, month=4, day=30))

    # Essen City on 21.06.2018
    # https://www.metaweather.com/de/648820/2018/6/20/
    temperature, weather_state = meta_weather_com.coordinates2weather(
        51.4109, 6.7828, date=datetime.datetime(year=2018, month=6, day=20, hour=20, minute=30)
    )
    temperature, weather_state = meta_weather_com.coordinates2weather(
        51.4109, 6.7828, date=datetime.datetime(year=2018, month=6, day=20, hour=13, minute=30)
    )
    temperature, weather_state = meta_weather_com.coordinates2weather(
        51.4109, 6.7828, date=datetime.datetime(year=2018, month=6, day=20, hour=4, minute=30)
    )
    temperature, weather_state = meta_weather_com.coordinates2weather(
        51.4109, 6.7828, date=datetime.datetime(year=2018, month=6, day=20, hour=20, minute=30)
    )
